

from numpy import array,  random, cos , sin , dot, pi

from tb.baseclass import Named

from ml.data import   TableSource
import csv


class RandomDataPart2D(Named):
    
    def init(self,
             left_bottom = (0.,0.),
             right_top   = (1.,1.),
             n_points    = 100 ,
             label       = 1,
             alpha       = 0.,
             ):
        
        self.data = []
        
        
        
    
    def build_data_part(self):
        
        a0 = scale_0   = self.right_top[0] - self.left_bottom[0]
        a1 = scale_1   = self.right_top[1] - self.left_bottom[1]
        b0 = offsett_0 = self.left_bottom[0]
        b1 = offsett_1 = self.left_bottom[1]
        
        self.rows = []
        
        for i in range(self.n_points):
            
            
            x0_raw = random.rand(1).ravel()[0] 
            x0     = b0 + a0 * x0_raw
            x1_raw = random.rand(1).ravel()[0] 
            x1     = b1 + a1 * x1_raw
            self.rows += [ [ x0, x1 , self.label ] ]
    
    def rotation(self):
        a = 2* pi * self.alpha/360.
        M = array( [[ cos(a), -sin(a)],
                    [ sin(a), cos(a)] ] )
        
        M_ = M.transpose()
        
        for i in range(self.n_points):
            p   = array(self.rows[i][:2])
            x   = p - array(self.left_bottom)
            x_  = dot(x, M_).transpose() + array(self.left_bottom)
            self.rows[i] = [ x_[0], x_[1] , self.label ] 

class RP(RandomDataPart2D):
    pass
         

            
class RandomData2D(TableSource):
    
    def init(self,
             parts = [RP(left_bottom = (1, 1.),
                         right_top   = (2, 3),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      ],
             seed = 0,
             separator = ','
             ):
        random.seed(seed)
        
    
    def file_to_rows(self):

        rows = []
        
        for part in self.parts:
            
            part.build_data_part()
            part.rotation()
            rows += part.rows
        
        file = open(self.source_abs, 'w')
        writer = csv.writer(file)
        writer.writerows(rows)
        file.close()  
        
        super(RandomData2D, self).file_to_rows()
        
 
        
        
          
  

           

class RD2D_A(RandomData2D):
    
    def init(self,
             parts = [ 
                      RP(left_bottom = (1, 1.),
                         right_top   = (2, 3),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (3, 1.),
                         right_top   = (4, 4),
                         n_points    = 60 ,
                         label       = 1,
                         alpha       = 0.
                         ),
                      RP(left_bottom = (5, 1.),
                         right_top   = (6, 5),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 0.
                         ),
                      RP(left_bottom = (7, 1.),
                         right_top   = (8, 6),
                         n_points    = 60 ,
                         label       = 1,
                         alpha       = -50.
                         ),
                      
                      ]
             ):
        pass
   

class RD2D_B(RandomData2D):
    
    def init(self,
             parts = [ 
                      RP(left_bottom = (1, 1.),
                         right_top   = (2, 3),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (3, 2.),
                         right_top   = (4, 4),
                         n_points    = 60 ,
                         label       = 1,
                         alpha       = 0.
                         ),
                      RP(left_bottom = (5, 3.),
                         right_top   = (6, 5),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 0.
                         ),
                      RP(left_bottom = (7, 4.),
                         right_top   = (8, 6),
                         n_points    = 60 ,
                         label       = 1,
                         alpha       = -50.
                         ),
                      
                      ]
             ):
        pass
   

class RD2D_C(RandomData2D):
    
    def init(self,
             parts = [ 
                      RP(left_bottom = (1, 2.),
                         right_top   = (2, 6),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (1, 1.),
                         right_top   = (5, 2),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (2.4, 3.),
                         right_top   = (4.4, 4),
                         n_points    = 60 ,
                         label       = 1,
                         alpha       = 10.
                         ),
                      
                      ]
             ):
        pass
   
class RD2D_D(RandomData2D):
    
    def init(self,
             parts = [ 
                      RP(left_bottom = (1, 2.),
                         right_top   = (2, 6),
                         n_points    = 40 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (1, 1.),
                         right_top   = (5, 2),
                         n_points    = 60 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (2.6, 3.2),
                         right_top   = (4, 4),
                         n_points    = 120 ,
                         label       = 1,
                         alpha       = 20.
                         ),
                      RP(left_bottom = (5, 1.),
                         right_top   = (6, 5),
                         n_points    = 40 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (1, 5.),
                         right_top   = (6, 5.7),
                         n_points    = 40 ,
                         label       = 0,
                         alpha       = 0.
                         ),
                      
                      
                      ]
             ):
        pass
   

class RD2D_E(RandomData2D):
    
    def init(self,
             parts = [ 
                      
                      RP(left_bottom = (1, 1.),
                         right_top   = (5, 2),
                         n_points    = 0 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (2.6, 3.2),
                         right_top   = (5.3, 4.5),
                         n_points    = 80 ,
                         label       = 1,
                         alpha       = 20.
                         ),
                      RP(left_bottom = (5, 1.),
                         right_top   = (6, 5),
                         n_points    = 40 ,
                         label       = 0,
                         alpha       = 10.
                         ),
                      RP(left_bottom = (1, 5.),
                         right_top   = (6, 5.7),
                         n_points    = 40 ,
                         label       = 0,
                         alpha       = 0.
                         ),
                      
                      
                      ]
             ):
        pass
   


class RD2D_F(RandomData2D):
    
    def init(self,
             parts = [ 
                      RP( 
                         left_bottom = (0,1.),
                         right_top   = (1.0,2.0),
                         n_points     = 40 ,
                         label       = 1,
                         alpha       = 30.
                         ),
                      RP(            
                         left_bottom = (2., 0.),
                         right_top   = (3., 3.),
                         n_points     = 40 ,
                         label       = 0,
                         alpha       = 45.,
                         ),
                      RP(
                         left_bottom = (4.,2.),
                         right_top   = (6, 4.),
                         n_points     = 40 ,
                         label       = 1,
                         ),
                      RP(
                         left_bottom = (4.5, -0.5),
                         right_top   = (5.5, 0.5),
                         n_points     = 20 ,
                         label       = 1,
                         ),
                      RP(
                         left_bottom = (7.5, 0),
                         right_top   = (9, 2),
                         n_points     = 20 ,
                         label       = 0,
                         )
                     ]):
        pass



       
            
if __name__ == '__main__':

    import pylab
    from tb.plotting import DataPlot
    def plot(self):
    
        X,Y = self.get_XY()
        
        dp = DataPlot(
                     show_grid     = True,
                     data_matrix   = X,
                     output_vector = Y, 
                     #title         = "candidate %d, value = %+f" %(i,v),
                     #lines         = lines[i], 
                     #line_colors   = ['b'],
                     #line_text     = [str(a) for a in I(lines[i])],
                         )
        dp.go()       
  
    
    rd = RD2D_E(seed=2)
    rd.build_data()
    #for k,v in rd.__dict__.items(): 
    #    print k
        
    X,Y = rd.get_XY()
    dp = DataPlot(
                     show_grid     = True,
                     data_matrix   = X,
                     output_vector = Y, 
                     #title         = "candidate %d, value = %+f" %(i,v),
                     #lines         = lines[i], 
                     #line_colors   = ['b'],
                     #line_text     = [str(a) for a in I(lines[i])],
                         )
    dp.go()     
   
    
    pylab.show()
    
    
    
