from numpy import dot, linalg, exp, shape, mean, diag ,repeat,\
                  vectorize, sqrt, where, ones, empty, reshape, nan, array, \
                  log2,median, random,max,min, argmax, argmin, zeros


from tb.baseclass import Named
from tb.scipytools import  tp, col, row, shapes
from tb.odict import OrderedDict
from ml.preprocessing import Standardization

class Kernel(Named):
    
    def init(self,
             type = 'dummy_kernel',
             c    = 0,
             g    = 1,
             sigma =1,
             fixed = False,
             
             ):
        pass

    def get_hash(self):
        out = self.get_name() + "__" + str(self.c)+ "__" + str(self.g) + "__" + str(self.sigma)  
        return out
    
    
    #def get_name(self):
    #    return self.type



class Poly(Kernel):

    def init(self,
                type='poly',
                c = 1.,
                g = None):
        
        if g is None:
            g = 1
            self.name = 'poly'
        else:
            self.name = 'poly' + str(int(g))


    def apply(self, x, x_):
        return ( dot(x, x_) + self.c )**self.g

        

class Linear(Poly):
    
    def init(self,
             type = 'linear',
             c    = 0.,
             g    = 1.
             ):
        assert c == 0.
        assert g == 1.
        
        self.name = 'linear' 
    

class Expl(Kernel):
    def init(self,
             type  = 'expl',
             sigma = 1.,
             ):
        pass

        self.name = 'expl' 
    
    def apply(self, x,x_):
        return exp( - abs(linalg.norm(x-x_) / ( self.sigma) ) )


 

class Gauss(Kernel):
    def init(self,
             type  = 'gauss',
             sigma = 1.,
             ):
        pass

        self.name = 'gauss' 
    
    def apply(self, x,x_):
        return exp( - (linalg.norm(x-x_)**2 / ( 2 * self.sigma) ) )



class GaussK(Kernel):
    def init(self,
             type  = 'gaussk',
             sigma = 1.,
             fixed = False,
             ):
        pass


    def get_name(self):
        if self.fixed:
            return "gauss_" + str(self.sigma)
        else:
            return "gauss_s"


    
    def apply(self, x,x_):
        t = dot(x, x_)
        a = 2**self.sigma
        return (a**t - 1)/(a-1)


    def apply_on_dpgm(self, K0):
        a = 2**self.sigma
        return (a**K0 - 1)/(a-1) 



class PolyK(Kernel):
    def init(self,
             type  = 'polyk',
             sigma = 1.,
             fixed = False,
             ):
        pass
    
    def get_name(self):
        if self.fixed:
            return "poly_" + str(self.g)
        else:
            return "poly_g"

    
    def apply(self, x,x_):
        t = dot(x, x_)
        a = self.g
        return t**a


    def apply_on_dpgm(self, K0):
        a = self.g
        return K0**a
        



def GK(type,**args):
    
    defaults = {
                'g'    : None, 
                'c'    : 1.,
                'sigma': 1.,
                'fixed': False
               }

    for k,v in defaults.items():
        if not args.has_key(k):
            args[k] = v
    
    kernel_type = type
    
    
    if   kernel_type == 'linear':   kernel = Linear()        
    elif kernel_type == 'poly':     kernel = Poly(**args)
    elif kernel_type == 'expl' :    kernel = Expl(**args)
    elif kernel_type == 'gauss':    kernel = Gauss(**args)
    elif kernel_type == 'gaussk':   kernel = GaussK(**args)
    elif kernel_type == 'polyk':    kernel = PolyK(**args)
    else:
        print('Wrong kernel type.')
        raise Exception()
 
    return kernel




class KernelTray(Named):

    def init(self,
             cache={},
             cache_cross ={},
             cache_cross_dp = {},
             kernel=None,
             data_tray=None,
             pruning_log_bound=None,
             use_cache = True,
             pp          = Standardization,
             #centralize_after_transform=False
             ):
        pass
        
        self.pp = pp()
        #if self.centralize_after_transform:
        #    M.X__mean       = col( mean(M.X__, axis=0) ) 
        #    Xn              = M.X__ - row(M.X__mean)
        #else:
        #    Xn              = M.X__
        
        
        

    def build_pp(self, A):
        X    = self.data_tray.get_X(A)
        self.pp.train(X)
        self.pp.A = A
        

    def query_cross(self,A,B, dp=False):
    
        def build_K(A,B):
            
            if not dp and self.kernel.has('apply_on_dpgm'):
                K = self.kernel.apply_on_dpgm( self.query_cross(A, B, dp=True) )
            else:
                XA    = self.data_tray.get_X(A)
                XB    = self.data_tray.get_X(B)
                XA_   = self.pp.apply(XA)
                XB_   = self.pp.apply(XB)
                
                if dp:    
                    K_0    = dot(XA_,XB_.transpose())
                    K_dA   = array([ linalg.norm(x) for x in XA_] ) 
                    K_dB   = array([ linalg.norm(x) for x in XB_] ) 
                    K_D    = dot(col(K_dA),row(K_dB))
                    #print "AAA", shapes(K_0, K_D), shapes(XA_, XB_), A == B
                    K_N    = K_0/K_D
                    K      = (K_N + 1)/2. 
                    eps    = 10**-10
                    assert all( (0 - eps) <= K.ravel() ) and\
                           all(K.ravel() <= (1 + eps) ) 
                else:
                    K = self.km(self.kernel.apply, XA_,XB_)
                    
            return K
        
        
        if dp:
            cache      = self.cache_cross_dp
            kernel_key = "dp" + self.pp.get_name() + str(self.pp.A)
        else:
            cache      = self.cache_cross
            kernel_key = self.kernel.get_hash() + self.pp.get_name() + str(self.pp.A)
        
        
        if cache.has_key(kernel_key) and self.use_cache:

            #self.send("get cahced ", self.kernel.get_hash())
            
            kernel_instance = cache[kernel_key]
            reuse           = False
                
            for index_key,K_ in kernel_instance.items():
                
                A_, B_  = index_key
                        
                if  set(A).issubset(set(A_)) and\
                    set(B).issubset(set(B_)):
                    
                    reuse = True
                    
                    iA = [ A_.index(a) for a in A]
                    iB = [ B_.index(b) for b in B]
                        
                    KM = K_[iA,:][:,iB]
                    break
                
            if not reuse:
                index_key    = ( tuple(A), tuple(B) )
                KM           = build_K(A,B)            
            
                kernel_instance[index_key] = KM    
        else:
            
            
            kernel_instance            = OrderedDict()
            
            index_key                  = ( tuple(A), tuple(B) )
            KM                         = \
            kernel_instance[index_key] = build_K(A,B)            
            cache[kernel_key]          = kernel_instance
        
        
        return KM
    

    
    def flash_cache(self):
        self.send('flash cache.')
        if self.cache_cross:            
            self.cache_cross.clear()
        self.cache_cross_dp.clear()
    
    
            
    def distances(self, S, T):
        
        #assert not set(S).intersection(T)
        
        K_ST    = self.query_cross(S,T)
        
        K_S  = self.query_cross(S,S)
        K_T  = self.query_cross(T,T)
        
        n = len(S)
        m = len(T)
        
        D    = nan * zeros((n,m)) 

    
        for s in range(n):
            for t in range(m):
                D[s,t] =  sqrt(K_S[s,s] + K_T[t,t] - 2 * K_ST[s,t]) 
                self.send("D:  %.3f = sqrt(%.3f + %.3f - 2*  %.3f )" %\
                          (D[s,t],K_S[s,s] , K_T[t,t] , K_ST[s,t]))
        return D
    



    def decomposition(self,M):
        """Computes eigenvalues decomposition of kernel matrix K.
        Rarely svd fails. Since we do not want this to break the runs we 
        retry with random noise added to K a fixed number of times.
        
        Returns UT,D,U, with orthonormal U. The eigenvectors are rows of U
        and columns of UT. D is one-dimensional.
        """
        
        K = M.K
        n = len(K)
                
        trials = 0
        success = False
        while not success and trials < 8:
            try:
                QT,D,Q  = linalg.svd(K)
                success = True
            except:
                self.send('SVD failed, retry with altered matrix.')
                K = K + random.rand(n,n) * 2*-20
        if not success:
            raise Exception()
        
        ## prune dimensions
        if not  self.pruning_log_bound is None:
            t  = 2**(mean(log2(D)) + self.pruning_log_bound)
            q  = sum( D >= t )
            D  = D[:q]
            Q  = Q[:q,:]
            QT = Q[:,:q]
            self.send("subspace dimensions :  %d/%d" % (q,n))
        else:
            q  = len(D)
        
        B  = dot(diag( D**(-1./2) ), Q ) 
        
        K_ = dot(tp(B),B)

        S  = sum(1/D)
        
        M.QT, M.D, M.Q, M.B, M.K_, M.S, M.n, M.q = QT,D,Q,B, K_, S, n,q 
         
        return M
    
    
    def raw_transform(self, X, kernel):
        K        = self.km(kernel.apply, X)            
        return K
    

    
    def km(self,f, v_1, v_2=None):
        
        def f_( x, x_ ):
            assert f(x,x) >= 0 and f(x_,x_) >= 0, str((f,x, x_))
            return f( x, x_ ) 
        
        if v_2 == None:
            v_2 = v_1
        
        return self.crosswise(f_, v_1, v_2)

    
    def crosswise(self,f_, v_1, v_2 ):
        """f_ is float valued"""
        n_1 = len(v_1)
        n_2 = len(v_2) 
        out = nan * empty((n_1,n_2), dtype='float32') 
        for i in range(n_1):
            for j in range(n_2):
                out[i,j] = f_(v_1[i],v_2[j])
        return out

    
    
    def pairwise(self,f, values):
        """f is float valued"""
        v  = values
        n = len(v)
        out = nan * empty((n,n), dtype='float32') 
        for i in range(n):
            for j in range(n):
                out[i,j] = f(v[i],v[j])
        return out
    
    
    def nkm(self,f, v_1, v_2=None):
        """normalized variant of crosswise"""
        def f_( x, x_ ):
        
            if f(x,x) > 0 and f(x_,x_) > 0:
                return f( x, x_ ) / sqrt( f(x,x) * f(x_, x_) )
            else:
                return f( x, x_ )* 0.
        if v_2 == None:
            v_2 = v_1
            
        return self.crosswise(f_, v_1, v_2)
    
    
    def normalize_km(self,km):
        
        n,n = shape(km)
        diagonal = sqrt(diag(km))
        diagonal = where(diagonal == 0, ones(shape(diagonal)), diagonal)
        assert all(diagonal)
        
        return km /\
               ( repeat(reshape(diagonal,(1,n)) ,n,axis=0 ) *\
                 repeat(reshape(diagonal,(n,1)) ,n,axis=1 ) )
    
    
    

