"""A libsvm wrapper that uses a hacked libsvm to access model alpha. """

from tb.baseclass import Named
from tb.scipytools import col
from tb.mixedtools import indexed

from ml.kernel    import GK, KernelTray
from ml.model import LLMM
from numpy import array, shape

import svm as libsvm

class LIBSVM(Named): 
     
    def init(self,
                data_tray   = None,
                kernel      = GK(type='linear'),
                C           = 1.,
                nu          = 0.1,
                svm_type    = 'C',
                kt          = None
        ):
        
        self.engine = libsvm
        
        libsvm.svm_parameter(kernel_type = self.engine.LINEAR)
        
        parameter   = self.engine.svm_parameter(kernel_type = self.engine.LINEAR)
            
        if   svm_type == 'C':
            parameter.svm_type = self.engine.C_SVC 
            parameter.C = C
        elif svm_type == 'nu':
            parameter.svm_type = self.engine.NU_SVC
            parameter.nu = nu
            self.send("nu: ", nu)
        else:
            self.bugger('Wrong svm_type.')
        
        ktype = kernel.type
        if  ktype == 'linear':    
            parameter.kernel_type = self.engine.LINEAR

        elif ktype == 'poly':
        
            parameter.kernel_type = self.engine.POLY
            parameter.coef0  = kernel.c
            parameter.gamma  = 1.
            parameter.degree = kernel.g
        
        elif 0:    
            if ktype == 'poly2':
            
                parameter.kernel_type = self.engine.POLY
                parameter.coef0  = kernel.c
                parameter.gamma  = 1.
                parameter.degree = 2
                
            elif ktype == 'poly3':
            
                parameter.kernel_type = self.engine.POLY
                parameter.coef0  = kernel.c
                parameter.gamma  = 1.
                parameter.degree = 3
                
                    
        elif ktype == 'gauss':
            
            parameter.kernel_type = self.engine.RBF
            parameter.gamma = 1./(2. * kernel.sigma)
        
        elif ktype == 'expl':
            
            parameter.kernel_type = self.engine.PRECOMPUTED
            #self.sigma = kernel.sigma
              
        else:
            self.bugger('Wrong kernel type: ' + ktype)
        
        self.parameter = parameter
        
    def set_C(self, C, n):
        
        self.parameter.C = C/float(n) 

    def set_sigma(self,sigma):
        self.sigma = sigma
        self.parameter.gamma = 1./(2. * sigma**2)

    def train(self, S=None, X=None, Y=None, sign=+1):
        """S are indices of the training examples"""
        
        
        if X is None:
            X,Y    = self.data_tray.get_XY(S)
        
        Y        = sign * Y 
        

        model = LLMM()
        model.sign=sign       
        
        model.sigma = self.kt.pp.sigma
        model.mu    = self.kt.pp.mu
        
        X_ =    self.kt.pp.apply(X)  
        
        problem            = self.engine.svm_problem(Y, X_)
        model.libsvm_model = self.engine.svm_model(problem, self.parameter)
        
        return model
    
    
    def predict(self, model, T=None, X=None, get_scores = False):
       
        if X is None:
            X_test            = self.data_tray.get_X(T)
        else:
            X_test = X
        
        X_test = self.kt.pp.apply(X_test)
    
        libsvm_model  = model.libsvm_model
        
        Y_hat             = array( [ libsvm_model.predict(x) for x in X_test]  )
        
        if get_scores:
            f_test = array( [ libsvm_model.predict_values(x)[(+1,-1)] for x in X_test])
            return Y_hat, f_test
        else:
            return Y_hat




if __name__ == '__main__':
    
    pass    
