"""A libsvm wrapper that uses a hacked libsvm to access model alpha. """

from tb.baseclass import Named
from tb.scipytools import col, a2sign
from tb.mixedtools import I

from ml.kernel    import GK, KernelTray
from ml.model import Model
from numpy import array, dot

import msvm as milk

class MILKSVM(Named): 
     
    def init(self,
                data_tray   = None,
                kernel     = GK(type='linear'),
                C           = 1.,
                engine      = milk.svm_learn_libsvm,
                kh        = KernelTray(),
                
                
        ):
        
        
        
        
        ktype = kernel.type
        if  ktype == 'linear':    
            rkernel = milk.polynomial_kernel(d=1.,c=0.)
            
        elif ktype == 'poly':
            rkernel = milk.polynomial_kernel(d=kernel.g, c=kernel.c)
            
            
        elif ktype == 'poly2':
            rkernel = milk.polynomial_kernel(d=2., c=0.)
                
        elif ktype == 'poly3':
            rkernel = milk.polynomial_kernel(d=3., c=0.)
                
        #elif ktype == 'gauss':
        #    rkernel = milk.rbf_kernel(sigma=kernel.sigma, beta=1.)
                        
        else:
            self.bugger('Wrong kernel: ' + ktype)
        
        self.rkernel = rkernel
        
    def set_C(self, C, n):
    
        self.C = C/float(n) 
        
    #def set_sigma(self,sigma):
    #    self.sigma = sigma
    #    self.rkernel.sigma =  2. * sigma

        
    def train(self, S=None, X=None, Y=None, sign=+1):
        """S are indices of the training examples"""
        
        model = Model()
        model.sign=sign  
        
        if X is None:
            X,Y    = self.data_tray.get_XY(S)
            model.S = S

        assert set(list(Y.ravel())) == set([-1,+1])
        
             
        model.build_scale(X)
        
        X_ = model.apply_scale(X)
        
            
        model.alpha,model.b = self.engine(X_, Y, self.rkernel, self.C)
           
        model.X_ = X_
        model.Y  = Y 
        return model
    
    
    def predict(self, model, T=None, X=None):
    
        M  = model
        
        if X is None:
            K_cross = self.kh.query_cross(T, M.S, self.rkernel, M)
            
        else:
            X_test  = X
            X_test  = M.apply_scale(X_test)
            K_cross = self.kh.km(self.rkernel, X_test, M.X_)
        
        G        = dot(K_cross, col(M.alpha) * col(M.Y) ).ravel()
        F        = G - M.b
        Y_hat    = a2sign(F)

        return Y_hat
         
        
        



if __name__ == '__main__':
    
    mm = MILKSVM()
    
    labels = array([-1, -1, +1, -1,+1])
    samples = array([[0, 0],
                     [0.3, 1.5], 
                     [1, 0.4], 
                     [1, 0.1], 
                     [-4, -4] ]
                    )
    model = mm.train(X=samples, Y=labels)  
    y_hat = mm.predict(model, X=[[0.5,1]])
    print y_hat
    print model.alpha
    
      
