from copy import deepcopy
from scipy import array, ones, shape, concatenate, dot, sum, linalg, rand, stats, \
    ravel, inf
from tb.baseclass import Named
#from tb.plotting import FunctionPlot
from tb.scipytools import tp, a2sign, row, col, select_rows
#import pylab

from openopt import NLP



class LC(Named):
    "Loss compound, base function for parts of a loss function"""

    def h_(self):
        """The base function of the compund."""
        self.abstract_method_bugger()

    def H(self):
        """The indefiniete integral of the base function h_."""
        self.abstract_method_bugger()

    def N(self):
        """normal distribution fab"""
        P      = stats.norm(0, 1) 
        p      = P.pdf
        F      = P.cdf
        
        def p_(x):
            return col(p(x.ravel()))

        def F_(x):
            return col(F(x.ravel()))
        
        return F_, p_ 


class Linear_LC(LC):
    """linear loss component"""
    
    def h_(self, m_):
        return  m_
    
    def h_dv_(self, m_):
        #print "SS", shape(ones(len(m_)))
        return  col( ones(len(m_) ))
        
        
    def H(self, u, a_, b_):
        
        F, p  = self.N()
        
        return  a_* - p(u)  + b_ * F(u) 

    def H_dv(self, u, a_, b_, b_du):

        F, p  = self.N()
        
        return    a_ * - u * - p(u) \
                + b_du * F(u) +  b_ * p(u)
    
    def H_llim(self, a_, b_):
        return 0
    
    def H_llim_dv(self,  a_, b_, b_du):
        return 0
    
    def H_rlim(self, a_, b_):
        return b_
    
    def H_rlim_dv(self,  a_, b_, b_du):
        return b_du


class Quadratic_LC(LC):
    """linear loss component"""
    
    def h_(self, m_):
        return  m_**2
    
    def h_dv_(self, m_):
        return  2 * m_
        
        
    def H(self, u, a_, b_):
        
        F, p  = self.N()

        H1  = a_**2 * ( F(u) -u * p(u) ) 
        H2  = 2 * a_ * b_ * ( - p(u) )
        H3  = b_**2 * F(u)
        
        return  H1 + H2 + H3 

    def H_dv(self, u, a_, b_, b_du):

        F, p  = self.N()
        
        H1_dv = a_**2 * ( p(u) - ( p(u) + u * -u * p(u) ))
        H2_dv = 2 * a_ * ( b_du * - p(u) + b_ * ( -u * - p(u) ) )
        H3_dv = b_du * 2 * b_ * F(u) + b_**2 * p(u)
        
        return H1_dv + H2_dv + H3_dv 
    
    def H_llim(self, a_, b_):
        return 0
    
    def H_llim_dv(self,  a_, b_, b_du):
        return 0
    
    def H_rlim(self, a_, b_):
        return a_**2  +  b_**2 
    
    def H_rlim_dv(self,  a_, b_, b_du):
        return b_du * 2 * b_

 


class R3SE_OB(Named):
    """Regularized, 3-sections error objective builder.
    """
    
    def init(self, 
                C_r=1., 
                C_n=1., 
                C_p=1., 
                
                symetric = False,
                
                smoothed = True,
                sigma = 0.1,
                s     = 1./100,
                # n
                r_n       = 0,
                q_n       = 0,
                
                lc_n_l    = Linear_LC(),
                a_n_l     = 1,
                q__n_l    = 0,
                c_n_l     = 0,
                d_n_l     = 0,

                lc_n_m    = Linear_LC(),                
                a_n_m     = 1,
                q__n_m    = 0,
                c_n_m     = 0,
                d_n_m     = 0,
                
                lc_n_r    = Linear_LC(),
                a_n_r     = 1,
                q__n_r    = 0,
                c_n_r     = 0,
                d_n_r     = 0,
                
                r_p       = 0,
                q_p       = 0,
                
                lc_p_l    = Linear_LC(),
                a_p_l     = 1,
                q__p_l    = 0,
                c_p_l     = 0,
                d_p_l     = 0,

                lc_p_m    = Linear_LC(),                
                a_p_m     = 1,
                q__p_m    = 0,
                c_p_m     = 0,
                d_p_m     = 0,
                
                lc_p_r    = Linear_LC(),
                a_p_r     = 1,
                q__p_r    = 0,
                c_p_r     = 0,
                d_p_r     = 0,
                    
                
                
                ):
        
        pass
    
    def set_data(self, X, Y):
        
        #assert set(tuple(Y)) == set([-1,1])
    
            
        #self.Xc   = concatenate( ( X, ones( (len(X),1) )  ), axis=1)
        self.X    = X
        Y         = col(Y)
        
        self.Xp   = select_rows( self.X, Y ==  1  )
        self.Xn   = select_rows( self.X, Y == -1  )
        
        
        self.nX   = col(array( [linalg.norm(x) for x in self.X] ))
        self.nXp  = select_rows( self.nX, Y ==  1  )
        self.nXn  = select_rows( self.nX, Y == -1  )

        self.Y    = Y
        self.Yn   = select_rows( Y, Y == -1 )
        self.Yp   = select_rows( Y, Y ==  1 )
        
        self.send()
        self.send('OB    :', self.get_name())
        self.send('C_r   :', self.C_r)
        self.send('C_n   :', self.C_n)
        self.send('C_p   :', self.C_p)
        self.send('sigma :', self.sigma)
        #self.send('slope :', self.s)




    def Q(self,v):
        v   = col(v) # any objective-part that is called from has to do this
        en  = sum(self.e(v,-1), axis=0)
        ep  = sum(self.e(v, 1), axis=0)
        r   = self.R(v)
        C_r = self.C_r
        C_n = self.C_n
        C_p = self.C_p

        return (C_r * r + C_n * en + C_p * ep).ravel()
    
    def Q_dv(self,v):
        v   = col(v)
        edvn  = self.e_dv_sum(v,-1)
        edvp  = self.e_dv_sum(v, 1)
        #print "SHA", shapes(edvn,edvp)
        rdv   = self.R_dv(v)
        C_r   = self.C_r
        C_n   = self.C_n
        C_p   = self.C_p
        
        return  C_r * rdv + C_n * edvn + C_p * edvp
 
    def Q_dv_flat(self,v):
        return ravel(self.Q_dv(v))


    def R(self,v):
        return self.sR(v) if self.smoothed else self.uR(v)
    
    def R_dv(self,v):
        return self.sR_dv(v) if self.smoothed else self.uR_dv(v)
 
 
    def e_sum(self,v,y):
        v = col(v)
        return sum(self.e(v,y), axis=0)

    def e_dv_sum(self,v,y):
        v = col(v)
        return row(sum(self.e_dv(v,y), axis=0))
    
    
    def e(self,v,y):

        el = self._e(v, y, 'l')
        em = self._e(v, y, 'm')
        er = self._e(v, y, 'r')
        
        return  el  +  em +  er 
    
    def _e(self,v,y,s):
        return self._se(v,y,s) if self.smoothed else self._ue(v,y,s)

    def e_dv(self,v,y):
        
        eldv = self._e_dv(v, y, 'l')
        emdv = self._e_dv(v, y, 'm')
        erdv = self._e_dv(v, y, 'r')
        
        return  eldv  +  emdv +  erdv 

    def _e_dv(self,v,y,s):
        return self._se_dv(v,y,s) if self.smoothed else self._ue_dv(v,y,s)
    
    
 
 
 
    ###unsmoothed R3SE OB"

    def uR(self,wb):
        """regularization term"""
        w = wb[:-1,0]
        return (1./2) * dot(tp(w),w)
                    
    def uR_dv(self,wb):
        w = wb[:-1,0]
        R_dw = row(w) 
        R_db = array([0], ndmin=2)
        return concatenate((R_dw,R_db), axis=1)


    def _ue(self,v,y,s):
        a, b, c, d, h_  = self.get_parameters( y, s, 'a','b','c','d','h_')
                
        m   = self.m(v, y)
        m_  = a * m + b
        
        cut = self.cut_factor(m,y,s)
        
        return cut * ( c * h_(m_) + d )
            
    def _ue_dv(self,v,y,s):
        a, b, c, d, hdv_  = self.get_parameters( y, s, 'a','b','c','d','h_dv_')
        
        m     = self.m(v, y)
        mdv   = self.m_dv(v, y)
        m_    = a * m + b
        
        cut = self.cut_factor(m,y,s)
        #print "MDU", shapes(mdv,cut,m_, hdv_(m_)) 
        return cut * ( c * a * mdv * hdv_(m_) )         



    ### smoothed
    def sR(self,wb):
        w = wb[:-1,0]
        return (1./2) * ( dot(tp(w),w) + len(w) * self.sigma**2 ).ravel()
  
    def sR_dv(self,wb):
        w = wb[:-1,0]
        R_dw = row(w) 
        R_db = array([0], ndmin=2)
        return concatenate((R_dw,R_db), axis=1)

    def _se(self,v,y,s):
        
        q, r, q_, r_, a, b, c, d, H, H_llim, H_rlim = self.get_parameters( y, s,
            'q', 'r', 'q_','r_','a','b','c','d','H','H_llim','H_rlim')
        
        m   = self.m(v, y)
        z   = self.z(y)
        
        qu  = (q - m)/ z
        ru  = (r - m)/ z
        
        a_  = a * z
        b_  = a * m + b
        
        
        #print v, a,b,c,d,a_,b_, qu, ru
        
        F     = stats.norm(0, 1).cdf
        
        if   s == 'l':
            E1  = H( qu, a_, b_) - H_llim( a_, b_)  
            E2  = F(qu) - F(-inf)
        elif s == 'm':
            E1  = H( ru, a_, b_) - H( qu, a_, b_)
            E2  = F(ru) - F(qu)
        elif s == 'r':
            E1  = H_rlim( a_, b_) - H ( ru, a_, b_)
            E2  = F(inf) - F(ru)

        return c * E1 + d * E2
        


    def _se_dv(self,v,y,s):
        
        q_, r_, a, b, c, d          = self.get_parameters( y, s, 'q_','r_','a','b','c','d')
        
        q, r, q_, r_, a, b, c, d, H_dv, H_llim_dv, H_rlim_dv = self.get_parameters( y, s,
            'q', 'r', 'q_','r_','a','b','c','d','H_dv','H_llim_dv','H_rlim_dv')
        
        m     = self.m(v, y)
        m_dv  = self.m_dv(v, y)
        z     = self.z(y)
        
        qu    = (q - m) / z
        ru    = (r - m) / z
        qu_dv = \
        ru_dv = \
        u_dv  = - m_dv / z
        #print "MDU", shapes(u_dv) 
        a_    = a * z
        b_    = a * m + b
        b_dv_ = a * m_dv
        b_du_ = a * -z 
        
        assert shape(a_) == shape(ru) 
        n = shape(m)[0]
        
        
        F    = stats.norm(0, 1).cdf
        F_dv = stats.norm(0, 1).pdf
    
        if   s == 'l':
            E_dv1 =  u_dv * ( H_dv(qu, a_, b_, b_du_) - H_llim_dv(a_, b_, b_du_) )
            E_dv2 =  u_dv * F_dv(qu) -  F_dv(-inf)
        elif s == 'm':
            E_dv1 =  u_dv  * ( H_dv(ru, a_, b_, b_du_) - H_dv(qu, a_, b_, b_du_) )
            E_dv2 =  u_dv * F_dv(ru) -  u_dv * F_dv(qu)
        elif s == 'r':
            E_dv1 =  u_dv * ( H_rlim_dv(a_, b_, b_du_) - H_dv(ru, a_, b_, b_du_))    
            E_dv2 =  F_dv(inf) -  u_dv * F_dv(ru)
        return c * E_dv1 + d * E_dv2 
    
  
    ### margins
    def m(self,v,y):
        
        X,Y = self.XY(y)
        #print shapes(Y,X,v)
        return Y * dot(X, v) 


    def m_dv(self,v,y):
        
        X,Y = self.XY(y)
             
        return Y * X 

    def z(self,y):
        sigma_0 = self.sigma
        nX  = self.nXn if y == -1 else self.nXp
        return sigma_0 * nX

 
  
    ### helper
    
    def XY(self,y):
        assert y in [-1,1]
        if y == -1:
            X = self.Xn
            Y = self.Yn
        else:
            X = self.Xp
            Y = self.Yp
        return X,Y    

    def cut_factor(self,m,y,s):
        q,r     = self.get_parameters( y, s,'q','r')

    
        if   s == 'l':
            out = ( m <= q )
        elif s == 'm':
            out = ( q < m ) * ( m < r)
        elif s == 'r':
            out = ( r <= m )
        return out

    
  
    def get_parameters(self, y, s, *names ):
        out = []
        for name in names:
            out += [ self.get_parameter(name, y, s) ]
        return out
    
    
    
    def get_parameter(self, name, y, s):
        assert s in ['l','m','r']
        assert y in [-1,1]
        label = 'n' if y == -1 else 'p'
        
        if self.symetric:
            label = 'n'
        
        if   name in ['a', 'c', 'd', 'q_']:
            attr_name = name + '_' + label + '_' + s
            out = self.__getattribute__(attr_name)
        elif name in ['q','r']:
            attr_name = name + '_' + label 
            out = self.__getattribute__(attr_name)
        elif name == 'r_':
            q,r,q_,a   = self.get_parameters(y, s, 'q', 'r','q_','a' )
            
            out  = q_ + a * (r-q)
            
        elif name == 'b':
            a  = self.get_parameter('a',  y, s)
            if s in ['l','m']:    
                q  = self.get_parameter('q',  y, s)
                q_ = self.get_parameter('q_', y, s)
            else:
                q  = self.get_parameter('r',  y, s)
                q_ = self.get_parameter('r_', y, s)
            out = - a*q + q_
        elif name in ['h_','h_dv_', 'H', 'H_dv',
                      'H_llim', 'H_llim_dv',
                      'H_rlim', 'H_rlim_dv', ]:
            attr_name = 'lc' + '_' + label + '_' + s            
            loss_compount = self.__getattribute__(attr_name)
            out = loss_compount.__getattribute__(name)
        else:
            print "GEP", name
            #bugger
            
        return out
    




 
class SVM_L1(R3SE_OB):
    def init(self,
                sigma     = 0.1,
                symetric  = True,
                # n
                q_n       =  1,
                r_n       =  1,
                
                lc_n_l    = Linear_LC(),
                a_n_l     = 1,
                q__n_l    = 0,
                c_n_l     = -1,
                d_n_l     = 0,
                
                lc_n_r    = Linear_LC(),
                a_n_r     = 1,
                q__n_r    = 0,
                c_n_r     = 0,
                d_n_r     = 0,
                       
                         
              ):
        pass
    


class SVM_L2(SVM_L1):
    def init(self,
                lc_n_l    = Quadratic_LC(),
                c_n_l     = 1,
              ):
        pass



class Q1_L1(R3SE_OB):
    def init(self,
                
                symmetric = False,
                
                # negatives
                q_n       =  1,
                r_n       =  1,
                
                lc_n_l    = Linear_LC(),
                a_n_l     = 1,
                q__n_l    = 0,
                c_n_l     = -1,
                d_n_l     = 0,
                
                lc_n_r    = Linear_LC(),
                a_n_r     = 1,
                q__n_r    = 0,
                c_n_r     = 0,
                d_n_r     = 0,
                
                
                # positives, to be separated
                
                q_p       =  1,
                r_p       =  1, 
                lc_p_l    = Linear_LC(),
                a_p_l     = 1,
                q__p_l    = 0,
                c_p_l     = -1./10**10,
                d_p_l     = 0,
                
                
                lc_p_r    = Linear_LC(),
                a_p_r     = 1,
                q__p_r    = 0,
                c_p_r     = -1,
                d_p_r     = -0,
              ):
        pass
      



class Q3(Q1_L1):
    def init(self,
                q_n       =  2,
                r_n       =  2,
                
                lc_n_l    = Quadratic_LC() ,
                c_n_l     = 1,
                d_n_l     = 0,
                q__n_l    = 0,
                
                lc_p_r    = Linear_LC(),
                a_p_r     = 1,
                q__p_r    = 0,
                c_p_r     = -1./10**5,
                d_p_r     = -1,
                
                ):
        pass



class Q1_L2(Q1_L1):
    def init(self,
                lc_n_l    = Quadratic_LC() ,
                c_n_l     = 1,
                d_n_l     = -1,
                q__n_l    = -1,
                ):
        pass


class Q2_L2(Q1_L2):
    def init(self,
                lc_p_r    = Linear_LC(),
                a_p_r     = 1,
                q__p_r    = 0,
                c_p_r     = -1./10**10,
                d_p_r     = -1,
                ):
        pass





                
def one_dim_cut_factory(f, v_0 , v_1):
    """for f:R^n -> B, a closure cut:R -> B
    with cut(r) = f(v_0 + r v_1_) is produced 
    where v_1_ = v_1/||v_1||"""
    
    v_0 = col(v_0)
    v_1 = col(v_1)
    
    assert shape(v_0) == shape(v_1)
    
    def cut(r):
        return f( v_0 + r * (1./ linalg.norm(v_1)) * v_1)
    
    return cut
    
def directional_derivative_factory(grad, v_0, v_1):
    
    grad_cut = one_dim_cut_factory(grad, v_0, v_1)

    def dd(r):
        return dot( grad_cut(r), col(v_1)/linalg.norm(v_1) )
    
    return dd

class OBTesting(Named):
    
    def init(self,
                n  = 4,
                d  = 6,
                ob = SVM_L1() ):
        pass
    
    def build_random_set(self, n=None, d=None):
        if n == None:
            n = self.n
        if d == None:
            d = self.d
        X = ( rand(n,d) - 0.5) * 2
        y = a2sign(rand(n,1) - 0.5 )
        self.ob.set_data(X,y)
    
    def plot_e(self, n=8,d=5):
        
        self.build_random_set(n, d)
                
        
        v_0 = self.random_vector(d, scale=2)
        
        
        v_1 = self.random_vector(d, scale=1)
        
        
    
        def e_sum(v):
            return (self.ob.e_sum(v,-1) + self.ob.e_sum(v,1)).ravel()
    
        def e_sum_dv(v):
            return (self.ob.e_dv_sum(v,-1) + self.ob.e_dv_sum(v,1)).ravel()
        
        e    = one_dim_cut_factory(e_sum, v_0 , v_1)
    
        e_dv = directional_derivative_factory(e_sum_dv, v_0, v_1)
    
        g = FunctionPlot( show_legend=False,
                          functions =  [ e , e_dv] ,
                          #x_lim=[0.2,1.2]
                         )
    
        g.go()
        pylab.show()
        
        
    
    def plot_m_e(self,y=1.):
    
        X = array( 1., ndmin=2 )
        Y = array( y,  ndmin=2 )
          
        self.ob.set_data(X,Y)
        
        def f_fab(ef,y):
            
            def e(v):
                return ef(v,y).ravel()
            
            return e
        
        uob = deepcopy(self.ob)
        sob = deepcopy(self.ob)
                
        uob.smoothed = False
        sob.smoothed = True
        
        se   = f_fab(sob.e_sum, y)
        ue   = f_fab(uob.e_sum, y)
        
        print "SMO", sob.smoothed, sob.sigma
        sedv = f_fab(sob.e_dv_sum,y)
        uedv = f_fab(uob.e_dv_sum,y)
            
        
        g = FunctionPlot( show_legend=True,
                          functions =  [ ue, se ],               
                          #x_lim=[0.2,1.2]
                          #x_interval=[-5,5]
                        )
        g.go()
        


    def check_gradient_e(self, n=20, d = 10):
        
        
        self.build_random_set(n,d)
            
        #v = self.random_vector(d, 100)
        e    = self.ob.e_sum
        e_   = self.ob.e_dv_sum
        def fn(v):
            return e(v,-1)
        def fn_(v):
            return e_(v,-1)
        
        def fp(v):
            return e(v,1) 
        def fp_(v):
            return e_(v,1)
        
        
        self._check_gradient( fn, fn_, d, 1 )
        self._check_gradient( fp, fp_, d, 1 )
    
    def _check_gradient(self, f, f_, d, n_checks):
        for i in range(n_checks):
            w = self.random_vector(d, 4)
            print w
            
            NLP(f, w , df=f_ ).checkdf()    
    

    def random_vector(self,d, scale=1):
        return (rand(d)-0.5)*scale*rand(d)
        
    
class XOB(R3SE_OB):
    def init(self,
                sigma     = 0.1,
                symetric  = True,
                # n
                q_n       =  1,
                r_n       =  1,
                
                lc_n_l    = Quadratic_LC(),
                a_n_l     = 1,
                q__n_l    = 0,
                c_n_l     = 1,
                d_n_l     = 0,
                
                lc_n_r    = Linear_LC(),
                a_n_r     = 1,
                q__n_r    = 0,
                c_n_r     = 0,
                d_n_r     = 0,
                    
                lc_p_l    = None,                
              ):
        pass
    


if __name__ == '__main__' and 1:
    
    from tb.plotting import FunctionPlot
    import pylab

    OB = SVM_L1(
                sigma=0.5,
                smoothed=1
                )
    OT = OBTesting(ob=OB)
    
    
    OT.check_gradient_e()
        
    OT.plot_m_e( y = 1. )
    OT.plot_e( n=6,d=4 )
    pylab.show()


   

            
        
                    
        
       
        

        
        
    
    