from numpy import dot, linalg, shape, mean, std, array, sqrt, nanmax, median,min,max
from tb.textprocessing import TextProcessing as TP
from tb.baseclass import Named
from tb.collections2 import sort_a_by_b
from tb.mixedtools import I
from tb.scipytools import nan_std


class PPMM(Named): 
    """Pre-Processing Model Maker"""
     
    def train(self,A):
        self.abstract_method_bugger()
        
    def apply(self, model,E):
        self.abstract_method_bugger()
        

class SP(PPMM):
    """
    Straight Preprocessing
    """
    
    def center(self,X,axis=0):
        return min(X,axis=axis)
    
    def size(self,X,axis=0):
        center = self.center(X, axis=axis)
        return nanmax(abs( X - center), axis=axis)
    
    def train(self,X):  
        
        mu      = self.center(X, axis=0)        
        sigma   = self.size(X,axis=0)
            
        for i in I(sigma):
            if sigma[i] == 0.:
                sigma[i] = 1. 
                
        self.mu    = mu
        self.sigma = sigma
        
        
        
    def apply(self,X):
        
        X2  =  (X - self.mu)/self.sigma 
        return X2


class S01(PPMM):
    
    def center(self,X,axis=0):
        return min(X,axis=axis)
    
    def size(self,X,axis=0):
        return nanmax(abs( X - self.center(X, axis=axis)),axis=axis)
    
    def train(self,X):  
        
        n, d    = shape(X)
        mu      = self.center(X, axis=0)        
        sigma   = self.size(X, axis=0)
            
        for i in I(sigma):
            if sigma[i] == 0.:
                sigma[i] = 1. 
                
        self.mu    = mu
        self.sigma = sigma
        self.n     = n
        self.d     = d
        
        
    def apply(self,X):
        
        X2  =  (X - self.mu)/self.sigma 
        return X2


class Spm1(S01):
    
    def center(self,X,axis=0):
        return mean(X,axis=axis)


class Standardization(SP):
    
    def center(self,X,axis=0):
        return mean(X,axis=axis)
        
    def size(self,X,axis=0):
        return nan_std(X,axis=axis)


class ScalingPM1MD(S01):
    
    def center(self,X,axis=0):
        return median(X,axis=axis)
 

class Norming(Named):
    def norm(self,x):
        n = linalg.norm(x)
        if n == 0:
            return x
        else:
            return x/n 
       

class N(PPMM, Norming):
    
    def train(self,X):  
        pass
            
    def apply(self,X):
        X1  =  array( [ self.norm(x) for x in X])
        return X1


class SN( Standardization,Norming):
    
            
    def apply(self,X):
        X1  =  (X - self.mu)/self.sigma 
        X2  =  array( [ self.norm(x) for x in X1])
        return X2
     

class SNN( Standardization,Norming):
            
    def apply(self,X):
        X1  =  (X - self.mu)/(sqrt(self.d)*self.sigma) 
        X2  =  array( [ self.norm(x) for x in X1] )
        return X2
     


class NormalizationPM1( Spm1,Norming):
    
            
    def apply(self,X):
        X2  =  (X - self.mu)/self.sigma 
        X1  =  array( [ self.norm(x) for x in X2])
        return X1


class S01N(Norming,S01):
            
    def apply(self,X):
        X2  =  (X - self.mu)/self.sigma 
        X1  =  array( [ self.norm(x) for x in X2])
        return X1



            
class PhiForText(Named):
    
    def vanilla(self,
                token_sequences,
                n_max_features = 5000):
        
        self.token_set      =  list(TP.get_terms(self.token_sequences))
            
            
        self.send("Number of terms:", len(self.token_set))
        TF    = TP.TF(token_sequences, self.token_set)
        self.idf   = TP.IDF(TF)    
            # simple feature selection: sorted by growing idf 
            # -> sorted by declining df, take first n_max_features
        self.token_set = sort_a_by_b(self.token_set, self.idf)[:self.n_max_features]
        self.idf       = sort_a_by_b(self.idf, self.idf)[:self.n_max_features]
        self.send("Number of terms used:", len(self.token_set))

    
    
    def apply(self,token_sequence):
        tf_vector = TP.tf(token_sequence, self.token_set)
        assert shape(tf_vector) == shape(self.idf)
        out =  tf_vector * self.idf
        return out
 
 
