import collections
from operator import itemgetter


def sort_a_by_b(a,b,reverse=False, out="a"):
    
    if not len(a) == len(b):
        raise Exception()
    z = map(None,a,b)
    
    sorted_pairs = sorted(z, key=itemgetter(1), reverse=reverse )
    
    if out == "a":
        return map(itemgetter(0), sorted_pairs)
    if out == "ab":
        #print sorted_pairs
        return sorted_pairs
    
 


        
def max_indices(b, k=1):
    
    a = range(len(b))
    
    sorted_indices = sort_a_by_b(a,b,out="a", reverse=True)
        
    return sorted_indices[0:k]


        
def frequencies(seq, mapping = lambda x: x, 
                sort_by_frequency=False, relativ=False):
    """Returns list of (value, frequency) pairs where frequency 
    is the number of occurances of value in seq after mapping 
    is applied to each entry. The pairs are sorted by value.
    
    If relativ == True, frequency is normalized. 
    If sort_by_frequency == True, the list is sorted by frequency.
    """ 
    
    if sort_by_frequency: 
        sort_key = itemgetter(1)
        reverse  = True
    else:
        sort_key = itemgetter(0)
        reverse  = False
        
    d = collections.defaultdict(lambda: 0)
    
    n = len(seq)
    
    for item in seq:
        if relativ:
            d[mapping(item)] += 1./n
        else:
            d[mapping(item)] += 1
            
    return sorted(d.iteritems(), key=sort_key, reverse=reverse)


def most_frequent_item(seq):
    if seq:
        return frequencies(seq,  sort_by_frequency=True)[0][0]
    else:
        return None
    
    
def normalize_frequecies(seq):
    z = float(sum(seq))
    return [ float(x)/z for x in seq ]

def get_x(function):
    return [float(x) for x,y in function]

def get_y(function):
    return [y for x,y in function]

def get_xy(function):
    return get_x(function), get_y(function)



class BetterList(list):
    """Adds function sorting that returns the permuation
    of the sorting along with the sorted list"""
     
    def __init__(self,x):
        list.__init__(self,x)
        self.i_list = [ (i,x[i]) for i in range(len(x)) ]

    def sorting(self, key=None, reverse=False):
        self.i_list.sort(cmp=self.compare, key= key,reverse=reverse)
        
    def get_sorting_permutation(self, key=None, reverse=False):
        self.sorting(key=key, reverse=reverse)
        return [ x[0] for x in self.i_list ]
       
    def get_sorted_list(self, key=None, reverse=False):
        self.sorting(key=key, reverse=reverse)
        return [ x[1] for x in self.i_list ]

    def compare(self,x,y):
        if type(x[1]) == type(float()):
            return self.float_cmp(x[1],y[1])
        else:
            return x[1].__cmp__(y[1])
    
    def float_cmp(self,x,y):
        if x < y:
            return -1
        elif x == y:
            return 0
        else:
            return 1

    def __setitem__(self,i,value):
        self.i_list[i][1] = value
        

    def get_sub_list(self, indices):
        return [ self.i_list[i][1] for i in indices ] 
            

class HighScoreList(object):
    """Does not use a search tree for keeping the ordered list, 
    e.g. is very slow for greater length (> 200)"""
    
    def __init__(self, length):
        self.length     = length
        self.hs_list    = []
        self.min_score  = None
        self.is_filled  = False
        self.indices    = range(self.length)
        
        
    def include(self, item, score):
 
        # self.min_score remain None until self.is_filled
        if score > self.min_score:
            
            if self.is_filled:
                
                for i in self.indices: 
                    if score > self.hs_list[i][0]:
                        self.hs_list.insert(i,(score, item))
                        break

                self.hs_list.pop()
                self.min_score = self.hs_list[-1][0]
                
            else:
                if self.hs_list == []:
                    self.hs_list = [ (score,item) ]
                else:
                    last = True
                    for i in range(len(self.hs_list)): 
                        if score > self.hs_list[i][0]:
                            self.hs_list.insert(i,(score, item))
                            last = False
                            break
                    if last:
                        self.hs_list.append((score, item))
                
                if len(self.hs_list) == self.length:
                    self.is_filled = True
                    self.min_score = self.hs_list[-1][0]
            
        
    def get_winners(self):
            
        return [ item for (score,item) in self.hs_list ]
    


if __name__ == '__main__':
    
    
    if 1:
        A = [1, 2, 3, 2, 5,5,3,5]
        B = [33, 2 ,2, 1,20]
        print most_frequent_item(A)
    if 0:
        import scipy, Clock
        Clock.tok()
        for j in range(10):
            l = 2**j
            print l
            H = HighScoreList(1)
            
            for i in  xrange(10):
                score = scipy.rand()
                pass
                H.include( score, i)
            Clock.tok()
            print H.hs_list
        
 
    if 0:    
        a =  [4, 3, 54, 23, 1 ,54, 0, 54,2] 
        c = [float(x) for x in a] 
        #b = better_list(c)
        
        