import os

from tb.fileaccess import get_lines, homedir
from tb.mixedtools import has_unique_value, transpose_list,indexed
from tb.baseclass  import Named

from tb.odict import OrderedDict 

from tb.collections2 import BetterList
from numpy import array

class Column(Named):
    
    def init(self,
             has_missing_values = False
             ):
        pass


class Row(Named):
    pass


class Table(Named):
    
    def init(self,
             separator       = None,
             raw_rows        = OrderedDict(),
             raw_columns     = OrderedDict(),
             rows            = OrderedDict(),
             columns         = OrderedDict(),
                
             has_column_haeder = False,
             has_row_haeder    = False,
             column_header     = [],
             row_header        = [],
             basedir         = None, ## if set, it is used as path prefix
             missing_value   = 'missing',
                
                ):
            pass

                
    def file_to_raw_table(self,path):
        
        if self.basedir:
            path = os.path.join(self.basedir,path)
        
        
        lines         = self.get_data_lines(path)
        
        splitted_lines = self.get_splitted_lines(lines)
        
        if self.has_column_haeder:
            self.column_header = splitted_lines.pop(0)
            
        if self.has_row_haeder:
            self.row_header = []
            for line in splitted_lines:
                self.header_rows += [ line.pop(0) ]
                
        for i,line in indexed(splitted_lines):        
            self.raw_rows[i] = line
        
        if self.raw_rows:
            a_row = self.raw_rows[0]
            for j,entry in indexed(a_row):
                column = Column()
                column.index = j
                if self.has_column_header:
                    column.name = self.column_header[j]
                else:
                    column.name = None
                self.columns += [ column ]
            
        
    def digest_raw_table(self):

        
        for j,entries in self.raw_columns.items():    
            ## > data
            column.entries  = entries
            column.values   = sorted(set(entries))
            self.make_type(column)
            max_entry_width = max([ len(str(entry)) for entry in column.values ])
            if self.column_header:
                column.width = max( len(str(column.name)), max_entry_width )
            else:
                column.width = max_entry_width
                
            
        if self.column_header:
            self.raw_rows.pop(0)
            
        for i,entries in self.raw_rows.items():
            self.rows[i] = row = Row()
            row.entries = entries
            
    
    def make_type(self,column):
        new_entries = []
        has_missing_values = False
        
        for i, entry in indexed(column.entries):
            new_entry    = self.numberize(entry)
            new_entries += [ new_entry ]
            if new_entry == self.missing_value:
                has_missing_values = True
        
        
        if has_missing_values:
            rest = [ e for e in new_entries 
                     if not e == self.missing_value ]
            
            if not any([ e is None for e in rest ]):                
                replacement = sum(rest)/len(rest)
                for i,entry in indexed(new_entry):
                    if entry == self.missing_value:
                        new_entry[i] = replacement
                
        if any([ e is None for e in new_entries ]):
            column.type = 'non-numeric'
        else:
            column.type = 'numeric'
            column.entries = new_entries
    
    
    def numberize(self,s):
        """tries to convert string s into a number"""
        if s == self.missing_value:
            return s
        
        if s.count(','):
            l,r = s.split(',')
            ln = self.numberize(l)
            rn = self.numberize(r)
            if type(ln) is int and type(rn) is int:
                s = '.'.join([str(ln),str(rn)])
                
        try:
            v = int(s)    
        except:
            try:
                v = float(s)
                

            except:
                v = None    
        return v
    
    
    def make_rows(self):
        if self.columns:
            n_rows = len(self.columns.values()[0].entries)
            for i in range(n_rows):
                row = Row()
                entries = [ column.entries[i] for column in self.columns.values() ]  
                row.entries = entries
                self.rows[i] = row
        else:
            print "VVV"
        
    def popc(self,index):
        out = self.columns.pop(index)
        for row in self.rows.values():
            row.entries.pop(index)
    
        return out
    
    def show_table(self):
        
        out = ''

        for i,row in self.rows.items():
            for j,column in self.columns.items():
                entry = row.entries[j]
                out += ' ' * (column.width - len(str(entry)) + 1) + str(entry)
                
            out += '\n'
        
        self.send(out)             
        
        
    def get_data_lines(self,file_path):
        
        def is_data_line(line):
            comment_starts =  ['%', '@', '//']
            return len(line) > 0 and\
                   not any([line.strip().startswith(s) 
                            for s in comment_starts ])
        
        lines_striped =  [ line.strip() for line in get_lines(file_path)]

        lines = [ line for line in lines_striped if is_data_line(line) ]
        return lines


    def get_splitted_lines(self, lines):
        """splits data lines of input file by separator
        which is automatically detected using an heuristic"""
        
        if self.separator is None:
            self.separator = self.detect_seperator(lines)
        else:
            assert type(self.separator) is type('')

        splitted_lines = [ line.split(self.separator) for line in lines ]
        return splitted_lines
 
    def detect_seperator(self,lines):
        
        
        chr2int = lambda x: [chr(i) for i in range(255)].index(x)
        
        favorate_separators = [chr2int(c) for c in [',', '\t', ';'] ]
        
        indices    = range(255)
        C_rows     = [ [  line.count(chr(i)) for i in indices ]  for line in lines]
        C_columns  = transpose_list(C_rows)  
        
        candidates = [ i for i in indices 
                          if has_unique_value(C_columns[i]) and C_columns[i][0] > 0  ]
        
    
        l = len(candidates)
        if l == 0:
            self.send('cant find seperator.',lb=2)
        if l == 1:
            separator = candidates[0]    
        if l > 1:
            # drop alphanumerics from candidate list
            candidates = [ c for c in candidates if not chr(c).isalnum()]  
            l          = len(candidates)
            if l == 0:
                self.send('there are at least two alphanumeric'+\
                          ' seperator candidates, but no other.')
            if l == 1:
                separator = candidates[0]
            
            if l > 1:
                # more than one remaining candidate: 
                # take the first of most frequent ones. If runner-up is exactly 
                # one count less, take runnner-up
                
                # sort candidates descending according to counts
                
                counts = BetterList([ C_columns[i][0] for i in candidates ])
                counts_sorting    = counts.get_sorting_permutation(reverse=True)
                counts_sorted     = counts.get_sorted_list(reverse=True)
                candidates_sorted = [candidates[i] for i in counts_sorting]
    
                self.send("Separator candidates: ", [chr(i) for i in candidates_sorted], \
                    ", Indices: ", candidates_sorted, 
                    ", Counts: ", counts_sorted)
                    
                # take most frequent candidate or runner up
                maximizer         = candidates_sorted[0]
                maximum           = counts_sorted[0]
                runner_up         = candidates_sorted[1]
                runner_up_count   = counts_sorted[1]
                 
                if maximum == runner_up_count + 1:
                    # some chars in each column, take runner up    
                    separator = runner_up
                else: 
                    # take most frequent one
                    separator = maximizer
                
                # prefer favorate separators regardless their counts
                favorate_separators.reverse()
                for c in favorate_separators:
                    if c in candidates_sorted:
                        separator = c
                        break
                             
        self.send("detected separator: '%s' (%d)" % (chr(separator), separator))             
        return chr(separator)


 
        
        


if __name__ == '__main__':
    
    path = 'acute/diagnosis.data'
    #path ='breast_cancer/breast_cancer.arff'
    T = Table(basedir = homedir() + 'code/data_sets_1/')
    
    T.file_to_raw_table(path)
    T.digest_raw_table()
    T.show_table()

    
    
   