#from enodes_sgo import *
from peewit.core import ENode

from project_conf import *

                   
from ml.evaluation import er

from tb.baseclass import Object
from tb.odict import OrderedDict
from tb.mixedtools import indexed, I
from tb.collections2 import sort_a_by_b
from tb.scipytools import shapes

class tuning(ENode):
    
    def descent(self, kernel_tray):
        
        return [ None ]
    
    def ascent(self,vcube,d_inputs):
        opt_cube = vcube.aggg([ ('argmin', 'c', 'kernel_parameter'),
                                  ('mean','sub_sample'),
                                 #('sel', 'eval_final', 'error_rate'),
                                ])
        
        opt_cube2 = vcube.aggg([ ('min', 'c', 'kernel_parameter'),
                                  ('mean','sub_sample'),
                                 #('sel', 'eval_final', 'error_rate'),
                                ])
    
        optimizer = opt_cube.cube.ravel()
        optimum   = opt_cube2.cube.ravel()[0]
                                     
        box = Object()  
        box.c  = optimizer[0]
        box.kp = optimizer[1]
        box.value = optimum
        
        kt = d_inputs.kernel_tray
        
        self.send("Optimal parameters:", box.c, box.kp, box.value, kt.kernel.get_name())
        
        kt.flash_cache()
        
        return box 
        

class target_instance(ENode):
    
    def descent(self, sample):
        
        return list(sample[1])

    def ascent(self, d_values, a_values):
        
        subs = OrderedDict()
        for i,x in indexed(d_values):
            
            subs[x] = a_values[i]
        
        
        return subs
        
        

class ib_problem(ENode):
    
    def init(self, k=32):
        pass
    
    def sub_problem(self, A,target_instance, kernel_tray):

        kt = kernel_tray
        Y  = kt.data_tray.get_Y(A)

        D   = kt.distances(A,[target_instance]).ravel()
        assert len(D) == len(A)
        print shapes(D,Y)
        AS  = sort_a_by_b(A,D)[:]
        YS  = sort_a_by_b(Y,D)[:]
        ASn = [ AS[i] for i in I(AS) if YS[i] == -1 ]
        ASp = [ AS[i] for i in I(AS) if YS[i] == +1 ]
        
        
        last_n = AS.index(ASn[self.k])
        last_p = AS.index(ASp[self.k])
        
        A_sub = AS[:max(last_n, last_p)]
        
        return (A_sub,[target_instance])
    
    def descent(self, target_instance, sample, kernel_tray):
        
        A = sample[0]

        return [ self.sub_problem(A,target_instance,kernel_tray) ]
    
    def ascent(self,vcube,d_inputs, d_values):
        opt_cube = vcube.aggg([   ('argmin', 'c', ),
                                  ('mean','sub_sample'),
                                ])
        
        opt_cube2 = vcube.aggg([  ('min', 'c', ),
                                  ('mean','sub_sample'),
                                ])
    
        optimizer = opt_cube.cube.ravel()
        optimum   = opt_cube2.cube.ravel()[0]
                                     
        box = Object()  
        box.A_sub    = d_values[0][0]
        box.E_target = d_values[0][1]
        box.c        = optimizer[0]
        box.cv_value = optimum
        
        kt = d_inputs.kernel_tray
        
        self.send("Optimal parameters:", box.c, box.cv_value, kt.kernel.get_name())
        
        kt.flash_cache()
        
        return box 


class ib_eval(ENode):

    def descent(self,target_instance, sample, mm_sym, kernel_tray ):
        
        Y = []
        Y_hat = []
        
        print target_instance.a_value
        
        for e,sub in target_instance.a_value.items():
            
            A      = sub.A_sub
            E      = sub.E_target
            
            assert E == [e], (E,[e])
            c      = sub.c
            
            mm_sym.set_C(2.**c, len(A))
            model  = mm_sym.train(A )    
            Y_hat += list(mm_sym.predict(model, E))
            Y     += list(kernel_tray.data_tray.get_Y(E))
            
        print "DDD" 
        print Y
        print Y_hat
        error_rate   =  er(Y, Y_hat)
            
        
        
        return [  error_rate ]
        
        
    def explicit_decent_value_names(self):
        
        return [ 'error_rate', ]
        



class leafc(ENode):

    def descent(self, mm_sym, data_tray, kernel, sub_sample, c ):
        
        A, E      = sub_sample
        
        mm_sym.set_C(2.**c, len(A))
        
        model  = mm_sym.train(A )    
        Y_hat  = mm_sym.predict(model, E)
        Y      = data_tray.get_Y(E)
        
        error_rate   =  er(Y, Y_hat)
        
        #self.send(kernel.type, mm_sym.C, kernel.sigma, error_rate )
        
        return [  error_rate ]
        
        
    def explicit_decent_value_names(self):
        
        return [ 'error_rate', ]
            

class leaf(ENode):

    def descent(self, mm_sym, data_tray, kernel, sub_sample, c , kernel_parameter):
        
        A, E      = sub_sample
        
        mm_sym.set_C(2.**c, len(A))
        
        
        if not kernel.fixed:       
            if kernel.type is 'gaussk':
                kernel.sigma = 2. ** kernel_parameter
            if kernel.type is 'polyk':
                kernel.g     = kernel_parameter + 5
        
    
        model  = mm_sym.train(A )    
        Y_hat  = mm_sym.predict(model, E)
        Y      = data_tray.get_Y(E)
        
        error_rate   =  er(Y, Y_hat)
        
        #self.send(kernel.type, mm_sym.C, kernel.sigma, error_rate )
        
        return [  error_rate ]
        
        
    def explicit_decent_value_names(self):
        
        return [ 'error_rate', ]
        


class leaf_final(ENode):

    def descent(self, mm_sym, data_tray, kernel, sample, tuning):
        
        A, E      = sample

        c  = tuning.a_value.c
        kp = tuning.a_value.kp
        
        mm_sym.set_C(2.** c, len(A))
        

        if not kernel.fixed:    
            if kernel.type is 'gaussk':
                kernel.sigma = 2. ** kp
            if kernel.type is 'polyk':
                kp = 5  +  kp
                kernel.g     =  kp 
        
        
        
        #self.clock("train")
        model  = mm_sym.train(A )
        #self.send_clock("train")
        
        #print model.beta, model.tau, mm_sym.get_name()
        
        
        #self.clock("predict")        
        Y_hat  = mm_sym.predict(model, E)
        #self.send_clock("predict")
        
        Y      = data_tray.get_Y(E)
        
        error_rate   =  er(Y, Y_hat)
        
        return [  error_rate,
                  tuning.a_value.value,
                  c,
                  kp ]
        
        
    def explicit_decent_value_names(self):
        
        return [ 'error_rate',
                 'vc_error',
                 'c_opt',
                 'kp_opt'  ]







