#from enodes_sgo import *
from project_conf import *
from peewit.core import ENode

from ml import data
from ml.sampling import SubSampling
from ml.kernel import GK, KernelTray
from ml.preprocessing import Standardization
from tb.scipytools import indices_equal, col
from ml.libsvm2 import  LIBSVM_BB

from mm_bc_so import MMBCSO
from so_bc import SVML1 
from numpy import arange, linalg



class svd_pruning(ENode):
    def descent(self):
        return [ 0,
                #-2,
                #-4
                 ]


class seed(ENode):
    def descent(self):
        return [ 
                 'seco',
               ]

class data_source(ENode):
    
    def init(self, build=True, store=False):
        pass
    
    def descent(self):
        
        data.DataSource.global_datadir = PK.abs_path('data_sets_1/')
        data.DataSource.global_build   = self.build
        data.DataSource.global_store   = self.store
        
        A = [
              data.sonar(), 
              data.ionosphere(),
              data.colic(), 
              data.heartstatlog(),
              data.breast_cancer(),
              data.clean1(),
              data.diabetes(),
              data.parkinsons(),
            ]
            
        return  A 



class sample(ENode):
    
    def init(self,
             n_folds=8,
             max_train_size=1024):
        pass
    
    def descent(self, data_source, seed):    
        
        #self.clock('dataset')
        self.send_clock('dataset')
        self.send("\nDataset:", data_source.get_name(),'\n')

        sampler = SubSampling(
                              sequence  = data_source.rows, 
                              mapping   = data_source.output_mapping, 
                              seed      = seed,
                              n_folds   = self.n_folds,
                              max_train_size = self.max_train_size,
                              )
        
        return sampler.get_cv_indices()


class data_tray(ENode):
        
    def descent(self, data_source, sample):
        
        tray = data.TableTray(source=data_source,
                              training_sample=sample[0])
            
        return [ tray ]



class kernel(ENode):
    
    def init(self,g=3.):
        pass    
    def descent(self):

        k_1 = GK(type='polyk', g=1,fixed=True)
        k_2 = GK(type='polyk', g=2,fixed=True)        
        k_3 = GK(type='polyk', g=3,fixed=True)
        
        k_4 = GK(type='polyk')
        k_5 = GK(type='gaussk')
        
        return  [ k_1,
                  k_2,
                  k_3,
                  #k_4,
                  #k_5 
                  ]
        

class pruning_log_bound(ENode):
    
    def init(self, B=0):
        pass
    
    def descent(self):
        return [ self.B ] 


class kernel_tray(ENode):
    
    
    def init(self, pp=Standardization):
        pass
    
    def descent(self, data_tray, sample, kernel ):
        
        kt = KernelTray( kernel = kernel,
                         pruning_log_bound = -4.,
                         data_tray = data_tray,
                         use_cache=1,
                         pp = self.pp )
        
        data_tray.kernel_handler = kt
        
        A = sample[0]
        
        kt.build_pp(A)
                                    
        return [ kt ]
 
 
class sub_sample(ENode):

        
    def descent(self, data_source, data_tray, sample, seed):    

        sampler = SubSampling(
                              sequence  = data_source.rows, 
                              restriction  = sample[0], 
                              mapping   = data_source.output_mapping,
                              seed      = seed,
                              n_folds   = 8,
                              max_train_size = 1000,
                              )
        
                
        return  sampler.get_cv_indices() 
 

class sigma_sequence(ENode):
    
    def descent(self):
        
        R = [ (1./8)**k for k in [0,1,2,3] ]
        #R = [ (1./8)**k for k in [0,1,2] ]
    
        return [ R ]
    
    
    
class c(ENode):
    
    def descent(self):
        return  list(arange(4,17))
             

class kernel_parameter(ENode):
    
    def descent(self, kernel):
        
        if kernel.fixed:
            return [ 1 ]
        else:
            return range(-4,8)



class balance_svm(ENode):
    
    def init(self, value=False):
        pass
    def descent(self):
        if self.value:
            return [self.value]
        else:
            return [False, True]

class mm_sym(ENode):

    def descent(self, data_tray, kernel_tray):
   
        libsvmbb  = LIBSVM_BB( data_tray   = data_tray,
                               kernel_tray  = kernel_tray,
                               balancing_weights = False,
                             )
        
        return [libsvmbb,
               ]





class model(ENode):

    def descent(self, mm_sym, data_tray, sample, c ):
        
        A, E      = sample
        
        
        mm_sym.set_C(2.**c, len(A))
                
        
        
        model  = mm_sym.train(A)
        
        beta_r = col(model.beta)
        
        print
        print 2**c
        print sum(abs(beta_r) > 10**-10)
        print linalg.norm(beta_r)
        print model.tau
        

        return [model]
        






