from peewit.core import ENode


from tb.scipytools import indices_equal, nan_mean
from tb.mixedtools import I
from tb.baseclass import Object
from ml.sampling import SubSampling
from ml.model import SOS, LLMM
from ml.evaluation import er, make_stats

from numpy import nan, inf, shape, sort,arange, sum
from mm_bc_so import MMBCSO
from so_bc import SVML1, SecoObjectiveA2pv1
from random import shuffle




class turbo(ENode):

    def descent(self):
        #return [1,0]
        return [ 0 ]
        
    
class cut_objective(ENode):

    def descent(self):
        
                     
        return [
                #SecoObjectiveA1,
                SecoObjectiveA2pv1,
               ]
        
class cut_engine(ENode):
    
    def init(self,
             dummy = False):
        pass
    
    def descent(self, cut_objective, kernel_tray,sigma_sequence, data_tray):
        
        tal_seco  = MMBCSO(data_tray       = data_tray,                   
                           kernel_tray     = kernel_tray,
                           
                           objective_class = cut_objective,
                           #weights         = [C, C_minus, C_plus ],
                           sigma_sequence  = sigma_sequence,
                           dummy  = self.dummy
                           )
        return [tal_seco ]




class cut_zero(ENode):
    
    def descent(self, sample, mm_sym, cut_engine, selection_criterion,data_tray):
        
        cut    = Object()
        model  = LLMM(dummy=True, maker=cut_engine)
        U,UE   = sample

        Y              = data_tray.get_Y(U)        
        Y_E            = model.maker.data_tray.get_Y(UE)        
        
        assert data_tray is model.maker.data_tray
        
        Y_hat, cut.m   = model.predict(U, more=1) 
        Y_hat_E        = model.predict(UE)    
        
        cut.U  = U
        cut.UE = UE
        
        cut.U_  = [ U[i]  for i in I(U)  if Y_hat[i]   == -1 ]
        cut.UE_ = [ UE[i] for i in I(UE) if Y_hat_E[i] == -1 ]
                
        cut.ct   = make_stats(U,  Y,   Y_hat,   model.sign)
        cut.ct_E = make_stats(UE, Y_E, Y_hat_E, model.sign)

        del model
        
        cut.c           = 0
        cut.c_plus      = 0
        cut.tries       = 0
        cut.n_ancestors = 0
        
        cut.zero      = True
        assert cut.U_ == U
        assert cut.UE_ == UE

        return [ cut ]

    def ascent(self, vcube, d_values, d_inputs):
        
        cut = d_values[0]
        
        oc1  = vcube.aggg([ ('argmin', 'c'),
                            ('mean',    'sub_sample'), ])

        oc2  = vcube.aggg([ ('min',   'c'),
                            ('mean',  'sub_sample'), ])

        c_residual_tuned      = oc1.cube.ravel()[0]
        cut.error_res_tuning  = oc2.cube.ravel()[0]
        cut.next_superfluous  = cut.error_res_tuning < 1./32
        
        U_  = cut.U_
        UE_ = cut.UE_
        d_inputs.mm_sym.set_C( 2**( c_residual_tuned ), len(U_) )

        residual        = cut.residual = Object()
        
        model_r      = d_inputs.mm_sym.train(U_)
        
        Y               = model_r.maker.data_tray.get_Y(U_)
        Y_hat           = model_r.predict( U_ )                
        Y_E             = model_r.maker.data_tray.get_Y(UE_)
        Y_hat_E         = model_r.predict( UE_ )
         
        residual.ct     = make_stats(U_,  Y,   Y_hat,   model_r.sign)
        residual.ct_E   = make_stats(UE_, Y_E, Y_hat_E, model_r.sign)
        
        del model_r
        
        cut.value   = d_inputs.selection_criterion[0](cut)
        cut.value_e = d_inputs.selection_criterion[1](cut)
                
        return cut



class cuts_fan(ENode):
    
    def init(self, r=8):
        pass
    
    def descent(self, kernel_tray):
        return range(self.r)

    def ascent(self, a_values, d_inputs):
        d_inputs.kernel_tray.flash_cache()
        return a_values[-1]

class cuts(ENode): 
     
    def descent(self, cuts, sample, cut_zero):
        
        return [ cuts['cuts_fan'] ]

    def initial_siblings(self, cuts, sample, cut_zero):

        cut = cut_zero.a_value
        sos = SOS()
        sos.comlete = False
        sos.sequence =[ cut ]
        
        return {'cuts_fan': sos }
 

    def prune_condition(self, cuts):
        
        return cuts['cuts_fan'].comlete
        

    def ascent_pruned(self, vcube, d_values):
        
        cuts = d_values[0]         
        self.svalue  = cuts
        
        return cuts
             
    def ascent(self, vcube, d_values):
        
        cuts = d_values[0] 
    
        oc    = vcube.aggg([('argminp','value',  'c_cand', 'c_cand_plus', 'sign_cand', 'tries')])
        cut   = oc.cube.ravel()[0]
        
        oc2   = vcube.aggg([('argminp','value_e','c_cand', 'c_cand_plus', 'sign_cand', 'tries')])
        cut2  = oc2.cube.ravel()[0]
        cut.value_e_best = cut2.value_e
        
        self.send('selected cut  c, c_p, try, value, value_e, value_e_best: ' +\
                  '%02d %02d %1d   %0.4f %0.4f %0.4f'% 
                  (cut.c, cut.c_plus, cut.tries, cut.value, cut.value_e, cut.value_e_best), lb=2 )
        
        if   cut.drop: ## low covereage or beggarly left-overs
            cuts.comlete = True
            self.send("selected cut is droped", lb=2 )
                                  
        else:
            cuts.sequence += [ cut ]
            self.send("cut added, now n_cuts=%d, uncovered examples before,now: %d %d" \
                      % (len(cuts.sequence), int(cut.ct.oo), int(cut.ct.on)), lb=2 )
            
            if cut.next_superfluous: ## in the sense the residual error is very low
                cuts.comlete = True
                self.send("next cut superfluous, error_res_e: %0.6f" % cut.error_res_tuning, lb=2)
            else:
                cuts.comlete = False
                self.send("go for another one", lb=2)
            
        self.svalue      = cuts
        
        return cuts 
    

class sign_cand(ENode):
    
    def descent(self):
        return [ -1 , +1 ]

class c_cand(ENode):
    
    def descent(self):
        
        return range(8,15)

class c_cand_plus(ENode):
    
    def descent(self):
        #return [-1]
        return [ -1, 0, 1 ] 
    
class tries(ENode):
    
    def init(self, n = 4):
        pass
    def descent(self):
        return  range(self.n) 
 
class candidate(cut_zero): ## inherit ascent()
       
    
    def descent(self, sample, data_tray, cut_engine,  c_cand_plus,  
                      cuts, sign_cand, c_cand, tries, mm_sym,
                      selection_criterion ):
        
        last   = cuts.sequence[-1]
                
        cut_engine.weights = [1., 2.**c_cand, 2.**(c_cand + c_cand_plus)]
        self.send('sign,weights,try:', sign_cand , cut_engine.weights, tries)
        
        cut         = Object()
        self.clock()
        model       = cut_engine.train(last.U_, sign=sign_cand)
        self.send_clock()

        cut.U  = U  = last.U_
        cut.UE = UE = last.UE_
        
        Y              = data_tray.get_Y(U)        
        Y_E            = data_tray.get_Y(UE)        
        
        Y_hat, cut.m   = model.predict(U, more=1) 
        Y_hat_E        = model.predict(UE)    
        
        cut.U_  = [ U[i]  for i in I(U)  if Y_hat[i]   == -1 ]
        cut.UE_ = [ UE[i] for i in I(UE) if Y_hat_E[i] == -1 ]
        
        ct   = cut.ct   = make_stats( U,  Y,   Y_hat,   model.sign)
        ct_E = cut.ct_E = make_stats( UE, Y_E, Y_hat_E, model.sign)
             
        del model 
        
        cut.c           = c_cand
        cut.c_plus      = c_cand_plus
        cut.tries       = tries
        cut.n_ancestors = len(cuts.sequence)
        
        ## drop ?
        cut.beggarly_leftovers = min(ct.nn, ct.pn) < 32 
        cut.low_coverage       = ct.op/ct.oo < 1./16   
        cut.drop               = (cut.low_coverage or cut.beggarly_leftovers)
        self.send("drop? nc bl: ", 
                  cut.drop, cut.low_coverage, cut.beggarly_leftovers, lb=1 ) 
        self.send("no,po,nn,pn,cov: (%d %d)  (%d %d) %0.2f" 
                  % (ct.no, ct.po, ct.nn, ct.pn, ct.op/ct.oo ), lb=2)
        
        return [ cut ]
    

    def ascent_pruned(self, vcube, d_values, d_inputs ):
        cut         = d_values[0]
        cut.value   = inf
        cut.value_e = inf
        cut.value_e_best = 0.
        cut.superfluess = None 
        
        return cut
    
    def prune_condition(self,d_values):
        cut  = self.d_values[0]
        return cut.drop  
    
 
class selection_criterion(ENode):

    def descent(self):
        
        def error_total_train(cut):
            realized_mass   = cut.ct.op/cut.ct.oo
            residual_mass   = cut.ct.on/cut.ct.oo
        
            residual_error  = cut.error_res_tuning
            realized_error  = cut.ct.np/cut.ct.op if cut.ct.op else 0.
            
            realized_error_test  = cut.ct_E.np/cut.ct_E.op if cut.ct_E.op else 0 
        
            self.send("cut_errors A,E:" ,realized_error, realized_error_test )
            add_on     = (cut.c)*10**-10 + (cut.c_plus)*10**-12

            #((realized_mass) * realized_error )   + \
            value  =  ((realized_mass) * realized_error_test )   + \
                      ((residual_mass) * residual_error )   + add_on
        
            self.send('anc:', cut.n_ancestors, 
                      'oo on:', int(cut.ct.oo), int(cut.ct.on), 
                      'c:',   cut.c,
                      'c_p:', cut.c_plus, lb=0 )
            
            self.send(" error: %.3f = %.2f * %.2f  +  %.2f * %.2f  + %.7f" %\
                       (value,  realized_mass, realized_error, 
                                residual_mass, residual_error , add_on), lb=2)
        
            return value

        def error_real_train(cut):
            ct = cut.ct
            realized_mass   = ct.op/ct.oo
        
            realized_error  = ct.np/ct.op if ct.op else 0.
        
            value  =  ((realized_mass) * realized_error )           
        
            self.send('anc:', cut.n_ancestors, 
                      'oo on:', int(ct.oo), int(ct.on), 
                      'c:', ct.c, lb=0 )
            self.send(" error: %.3f = %.2f * %.2f  " %\
                       (value,  realized_mass, realized_error, 
                                ), lb=2)        
            return value

        
        def error_total_test(cut):
            ct_E = cut.ct_E
            realized_mass   = ct_E.op/ct_E.oo if ct_E.oo else nan
            residual_mass   = ct_E.on/ct_E.oo if ct_E.oo else nan
        
            realized_error  = ct_E.np/ct_E.op if ct_E.op else 0 
            
            ct_r_E = cut.residual.ct_E
            
            residual_error  =  (ct_r_E.np + ct_r_E.pn)/(ct_r_E.oo) if (ct_r_E.oo) else nan
            
            assert ct_E.on == ct_r_E.oo
        
            value  =  ((realized_mass) * realized_error )   +\
                      ((residual_mass) * residual_error )        
        
            return value
        
        return [ #(error_real_train,error_total_test), 
                 (error_total_train,error_total_test),
                 #(error_total_test, error_total_test)
               ]

class sub_sample(ENode):

    def descent(self, data_source, data_tray, cut_zero):    

        
        sampler = SubSampling(
                              sequence  = data_source.rows, 
                              restriction  = cut_zero.U_, 
                              mapping   = data_source.output_mapping,
                              seed      = 'ke',
                              n_folds   = 8,
                              max_train_size = 1000,
                              )
        
                
        return  sampler.get_cv_indices() 

class eval_sym(ENode):

    def descent(self, mm_sym, data_tray, sub_sample, c):
        
        A,B = sub_sample
        
        assert A and B, (A,B)
        assert data_tray.n_distinct_outputs(A + B) == 2, (A,B)
            
        mm_sym.set_C(2. ** c, len(A))
            
        final_model  = mm_sym.train( A, sign=+1 )

        Y      = data_tray.get_Y(B)                  
        Y_hat  = mm_sym.predict(final_model, B)
        
        error_rate   =  er(Y,Y_hat)

        return [  error_rate ]
        

  
        
class eval_sequence(ENode):

    def descent(self, cuts_fan, sample):
        
        #A             = sample[0]
        E             = sample[1]
        
        cuts     = cuts_fan.a_value.sequence
        last     = cuts[-1]
        pcuts    = cuts[1:]
        for cut in pcuts:
            ct_E = cut.ct_E
            cut.error_E     = cut.ct_E.np/ct_E.op if ct_E.op else 0.
            cut.fraction_E  = cut.ct_E.op/len(E)

        error_cut   = sum( [ c.fraction_E * c.error_E for c in pcuts ] )


    
        #print last.__dict__.keys()
        residual  = last.residual
        ct_r      = residual.ct
        ct_r_E    = residual.ct_E

        error_res  = (ct_r_E.np + ct_r_E.pn)/(ct_r_E.oo) if (ct_r_E.oo) else nan 
        min_frac   = min(ct_r.no,ct_r.po)#/(ct_r.oo)
        min_frac_E = min(ct_r_E.no,ct_r_E.po)#/(ct_r_E.oo)

            
        UE = last.UE_
            
        CE = [ e for e in E if not e in UE ]
        
            
        n_E     = 1. * len(E)
        n_C     = 1. * len(CE)
        n_U     = 1. * len(UE)
        assert n_U
                
        n_cuts       = len(cuts) * 1. 
        fraction_cut = n_C/n_E
        fraction_res = n_U/n_E
        
        error_total = fraction_cut * error_cut + fraction_res * error_res
        
        first_residual = cuts[0].residual
        ct_fr_E        = first_residual.ct_E
        error_plain    = (ct_fr_E.np + ct_fr_E.pn)/(ct_fr_E.oo)   
        
        
                
        return [
                error_plain,
                error_total,
                fraction_cut,
                error_cut,
                fraction_res,
                error_res,
                n_cuts,
                min_frac, 
                min_frac_E,
                  
               ]

        
    def explicit_decent_value_names(self):
        
        return [               
                'error_plain',
                'error_total',
                'fraction_cut',
                'error_cut',
                'fraction_res',
                'error_res',
                'n_cuts',
                'min_res', 
                'min_res_E',
               ]



class cut_one(cut_zero):
    
    def descent(self, sample, mm_sym, cut_engine, data_tray, selection_criterion):
        
        cut    = Object()
        cut.c           = 0
        cut.c_plus           = 0
        cut.n_ancestors = 0
        assert cut_engine.dummy
        
            
        U,UE   = sample

        Y              = data_tray.get_Y(U)        
        Y_E            = data_tray.get_Y(UE)        
        
        on = 0
        oo = len(U)
        nn = 0
        pn = 0
        count = 0
        while  nn < 32 or pn < 32 or on > oo * (1./ 2 + count * 1./1000) :  
            
            model      = cut_engine.train(U)
        
            Y_hat, cut.m   = model.predict(U, more=1) 
            Y_hat_E        = model.predict(UE)    
        
            cut.U_  = [ U[i]  for i in I(U)  if Y_hat[i]   == -1 ]
            cut.UE_ = [ UE[i] for i in I(UE) if Y_hat_E[i] == -1 ]
                
            cut.ct   = make_stats(U,  Y,   Y_hat,   model.sign)
            cut.ct_E = make_stats(UE, Y_E, Y_hat_E, model.sign)

            on = cut.ct.on
            nn = cut.ct.nn
            pn = cut.ct.pn
            oo = cut.ct.oo
            
            self.send("oo, on, nn, pn", oo, on, nn, pn)
            
        self.send('cut found', lb=2)

        del model
        
        return [ cut ]


class rand_sample_model(ENode):
    
    def descent(self, cut_one, sample, mm_sym, data_tray):
                
        nn = int(cut_one.a_value.ct.nn)
        pn = int(cut_one.a_value.ct.pn)
        
        A,E  = sample

        Y   = data_tray.get_Y(A)
        
        A_n = [ A[i] for i in indices_equal(Y, -1) ]
        A_p = [ A[i] for i in indices_equal(Y, +1) ]
        
        assert len(A_n) + len(A_p) == len(A)
        
        shuffle(A_n)
        shuffle(A_p)
        A_n_sub = A_n[:nn]
        A_p_sub = A_p[:pn]
        A_sub   = A_n_sub + A_p_sub
        shuffle(A_sub)
        
        assert len(A_sub) == nn + pn
        
        out = Object()
        out.U_  = out.A_sub = A_sub
        out.UE_ = out.E = E
    
        return [ out  ]

    def ascent(self, vcube, d_inputs, d_values):
        
        oc1  = vcube.aggg([ ('argmin', 'c'),
                            ('mean',    'sub_sample'), ])

        
        c_tuned = oc1.cube.ravel()[0]
        
        rmodel = d_values[0]
        
        A_sub  = rmodel.A_sub
        E      = rmodel.E
    
        d_inputs.mm_sym.set_C(2. ** (c_tuned), len(A_sub))

        model_sub = d_inputs.mm_sym.train(A_sub)

        Y     = d_inputs.data_tray.get_Y(E)                          
        Y_hat = d_inputs.mm_sym.predict(model_sub,E)
        
        out = Object()
        
        out.ct_E = make_stats(E, Y, Y_hat, model_sub.sign)
        out.A_sub   = A_sub
        out.E       = E
        
        return out


class plain_model(rand_sample_model):
    
    def descent(self, cut_one, sample, mm_sym, data_tray):
                
        
        out = Object()
        out.A_sub, out.E = out.U_, out.UE_ = sample
        
        return [ out  ]




class eval_rcrs(ENode):

    def descent(self, cut_one, rand_sample_model, plain_model, data_tray, sample):
        
        #A             = sample[0]
        
        cut  = cut_one.a_value
        ct_E = cut.ct_E
        error_cut     = ct_E.np/ct_E.op if ct_E.op else 0.
        fraction_cut  = ct_E.op/ct_E.oo 
        fraction_res  = ct_E.on/ct_E.oo
        
        
        residual   = cut.residual
        ct_r_E     = residual.ct_E
        error_res  = ct_r_E.error
        
        
        ct_rs_E      = rand_sample_model.a_value.ct_E
        error_sub    = ct_rs_E.error  
        
        ct_pl_E      = plain_model.a_value.ct_E
        error_plain  = ct_pl_E.error  
        
        
                
        return [
                error_plain,
                error_res,
                error_sub,
                error_cut,
                fraction_res,
                ]
        
    def explicit_decent_value_names(self):
        
        return [               
                'error_plain',
                'error_res',
                'error_sub',
                'error_cut',
                'fraction_res',

               ]




class sub_problems(ENode):
    
    def descent(self, plain_model, cut_one, rand_sample_model):
        return [ ( plain_model.a_value.A_sub,       plain_model.a_value.E ),
                 ( rand_sample_model.a_value.A_sub, rand_sample_model.a_value.E ),
                 ( cut_one.a_value.U_,              cut_one.a_value.UE_ ),
                 
               ]
    
    def explicit_decent_value_names(self):
        
        return [               
                'plain',
                'rand_sub',
                'one_cut',
               ]



class cross_distances(ENode):
    
    def init(self,k_max=32):
        pass

    def descent(self,sub_problems, kernel_tray):
        
        A,E  = sub_problems
        D    = kernel_tray.distances(E,A)


        Ds   = sort(D,axis=1)

        
        
        Dsk  = Ds[:,:self.k_max]
        
        #print Dsk
        
        return [ Dsk ]
        

class k(ENode):
    
    def descent(self, cross_distances):
        n,m = shape(cross_distances)
        return range(1,m)

    
class distance(ENode):
    def descent(self):
        pass
        return list(arange(0.2,1.44,0.02))

 
    
    
class reached_fraction(ENode):
    
    def descent(self, cross_distances, distance):
        
        D   = cross_distances 
        #D_E = nan_mean(D[:,:k], axis=1)

        R = sum( (distance-0.02 <= D) * (D < distance)  , axis=1)/ (1. * shape(D)[1])
        
        #R = sum((D < distance)  , axis=1)/ (1. * shape(D)[1])

        
        h = nan_mean(R,axis=0) 
        
        
        return [ h ]
        
    def explicit_decent_value_names(self):
        
        return [               
                r'$P(D\leq d)$',
               ]

