from numpy import array, dot, linalg, random,  diag, ones
from ml.artificial_data import RandomData2D
from ml.model import LLMM
from ml.optimization import fmin_bfgs
from tb.baseclass import Named
from tb.mixedtools import stems_from, indexed
from tb.scipytools import a2sign, col, append_value, tp



class MMBCSO(Named):
    """Model Maker for Binary Classification trough Smoothed Objectives.
    The class provides training for binary classification with an 
    l_2-C-SVM objective defined through the module so_bc.     
    """
    
    def init(self,                
                kernel_tray     = None,
                data_tray       = None,
                objective_class = None,
                objective_args  = {},
                weights         = None,
                sigma_sequence  = [0.1, 0.01, 0.0001],
                epsilon         = 10**-5,
                apply_BR        = False,
                dummy           = False,
                C               = 1.
                
                ):
        """
        @param kernel: kernel object as provided by the function GK
            from the module ml.kernel
        """
        pass
    
        
    
    def get_name(self):
        name = self.__class__.__name__ + '.' + self.objective_class.__name__ 
        if self.dummy:
            name += "_dummy"
        return name
    
             

    def set_C(self, C, n):
        
        self.C = C
        dummy_objective = self.objective_class(apply_BR=False)
        n_losses        = len(dummy_objective.losses)
        n_regulizers    = len(dummy_objective.regulizers)
        self.weights    = [ 1. / C ] * n_regulizers + [ 1. ] * n_losses
   



    def train( self, A=None, sign  = +1,  v_0   = None,):
        
        kt = self.kernel_tray
        
        M   = BCSOModel(maker=self)
        
        
        M.U = A
        M.X = self.data_tray.get_X(A) 
        M.Y = self.data_tray.get_Y(A) 
        M.sign = sign  
        
        M.sigma = kt.pp.sigma
        M.mu    = kt.pp.mu
            
     
        M.K   = kt.query_cross(A, A)
        M     = kt.decomposition(M)
        
        
        BL    = self.get_BL(M)
        
        Phi   = tp(dot(BL,M.K))
        M.Phi = Phi
        
        M.R   = max( [ linalg.norm(x) for x in Phi])
        #self.send("n,d,R: ", shape(Phi), M.R)
        #self.send(Phi[0])
        
        Phi_c = append_value(Phi, c = -1.)
    #        self.send(Phi_c[0:3])
        #self.send("PPC ", shapes(Phi, Phi_c, BL, M.K) )
        vb    = (random.rand(len(Phi_c[0])) - 0.5) * 16 \
                 if v_0 == None else v_0
             
        values = []
        self.send('q-values: ',lb=0)
   
        if not self.dummy:
                        
            for i, sigma in indexed(self.sigma_sequence):
                eps_scale = linalg.norm(vb)
                eps = eps_scale * 10**(-(1+i))
                #self.send("eps:", eps)
                objective  = self.objective_class(X     = Phi_c,
                                                  Y     = sign * M.Y,
                                                  sigma = sigma,
                                                  weights = self.weights,
                                                  M     = M,
                                                  dv_2  = False,
                                                  **self.objective_args
                                                  )
                # fmin_bfgs #fmin_ncg,
                opt_output = fmin_bfgs(         objective.Q, vb, objective.Q_dv, 
                                                #fhess=objective.Q_dv_2,
                                                maxiter     = 100, 
                                                retall      = True,
                                                full_output = True,
                                                disp        = 0,
                                                gtol        = eps#self.epsilon
                                              )
                
                del objective
                
                vb, value = opt_output[:2]
              
                try:
                    value = value[0]
                except:
                    pass
                values += [ value ]
                
                self.send(value,lb=0)
                
                if vb[-1] > 10**40:
                    self.send('break iterative opitmization, b = %f' % vb[-1] )
                    break
            
        else:
            value = 412.
            values += [ value ]
            
        self.send()
        #self.send('model_s: ', vb[:4], vb[-3:])
        #self.send('betas small/all: ', sum( abs(vb[:-1]) < 10**-3), len(vb) - 1,lb=2)  
        
        M.beta, M.tau  = vb[:-1], vb[-1]
        self.send("beta[:5],tau:", M.beta[:5], M.tau)
        #self.send('sigmas:  ',  [ '%.4f' % x for x in self.sigma_sequence])
        #self.send('minimas: ',  [ '%.4f' % x for x in values], '\n')
        M.value = value
        
        ## for plotting
        if stems_from(self.data_tray.source, RandomData2D):
            
            self.store_unit_images(M)

               
        return M
        

    def predict( self, M, E, more= False):
        
        Y_test = self.data_tray.get_Y(E)
        
        kt = self.kernel_tray
        
        kt.pp.mu    = M.mu
        kt.pp.sigma = M.sigma
        

        K_cross  = kt.query_cross(M.U,E)

        BL        = self.get_BL(M)

        Phi_test  = dot(BL, K_cross).transpose()
                
        g          = dot(Phi_test, col(M.beta)).ravel()
        f          = g - M.tau
        y_hat      = a2sign(f)
        m          = f * Y_test
        gm         = f / linalg.norm(M.beta) 
    
        return (y_hat, m) if more else y_hat
    
 
    def get_BL(self,M):
        if self.apply_BR:
            BL = diag(ones(M.n))
        else:
            BL = M.B
        return BL
 
    def store_unit_images(self,M):
        
        
        X12    = array([[1.,0],[0.,1]]) 
        X_test = X12
        
        #X_test = M.apply_norm(X_test)
        
        X_train = M.apply_scale(M.X)
        
        K_cross  = self.data_tray.kernel_tray.km(self.kernel.apply, 
                                                    X_train, X_test)
        
        M.E12_    = dot(M.B, K_cross).transpose()[:,:2]
        


class BCSOModel(LLMM):
    
    def init(self,
             beta  = None,
             tau   = None,
             X_ = None,
             h_ = None,
             K  = None,
             B  = None,
             X  = None,
             Y  = None,
             D  = None,
             U  = None,
             UT = None,
             sign = None,
             ):
        pass


  
    
