"""Global setting of the project may be define here.

The settings made here concern the names and paths for experiment and result cubes.
They are used in order to identify and address experiments and results.

Directories are given as relative paths in posix-notation but before they are applied 
they translated into the paths notation of the given os. That is, when we use those strings
to set the paths of an experiment, the strings a taken as I{paths-descriptors} from which the 
actual path string is derived. 
"""

from peewit.paths import PathKeeper 
from peewit.parallelization import SGE_Launcher, OPBS_Launcher
from peewit.archiving import Archivist



## cares for paths and name parts
PK = PathKeeper() 

## an identifier of your working group
## peewit assumes unique host name within your group 
PK.group = 'ke_tud_de'

##  Base_dir is given relative to your home directory. 
##  All paths set here are relativ to that base dir or relative to a
##  dir below. All paths are given as _path-descriptors_ using 
##  posix notation. peewit cares for translation to windows notation. 
PK.base_dir    = 'code/'  
                            
## project directory relative to base_dir
PK.project_dir = 'projects/mdsc_sgo/'

### python-path management

##  All paths will be added on the fly all but path to peewit itself
##  which is needed in advance. Paths are given relative to base_dir. 
##  plain user space paths
PK.python_paths    = [ 'modules_homebrewed/',
                       'modules_plain/openopt/' ]

## some paths may depend on architecture, 32 versus 64 bit
PK.python_paths_32 = ['modules_extended_32/libsvm-2.86/python/', ]
PK.python_paths_64 = ['modules_extended_64/libsvm-2.86/python/', ]

## others may not belong to user space but are from outer  
## space what also demands extra treatment. Those can also 
## be given as absolute paths. Keep in mind that
## "outer space, its not a better place, Grace"
PK.python_paths_os = [] 

## non-code paths needed by the experiments of the project
## and peewit or some services of peewit should now about
PK.other_paths = ['data_sets_1/']


### archiving settings

AR = Archivist()

#AR.other_repos = []
AR.exclude_from_archive     = []

AR.repos_passive_integration = []

AR.repos_active_integration  = []
# + PK.python_paths_32 + PK.python_paths_64 + PK.other_paths


AR.omit_pyd_files = True
## for peewit developers
AR.omit_peewit_restore = True

### settings for parallelization


class Launcher_Frankfurt(OPBS_Launcher):
   
    def init(self,
        remote_head              = 'hhlr.csc.uni-frankfurt.de',
        remote_head_architecture = '64bit',
        remote_user              = 'weizsaecker',
        remote_home_dir          = '/home/darmstadt/weizsaecker'
             ): 
        pass


class Launcher_E3(SGE_Launcher):
   
    def init(self,
        remote_head              = 'cluster',
        remote_head_architecture = '32bit',
        remote_user              = 'lw',
        remote_home_dir          = '/nis/home/lw/',
        queue                    = 'all.q@@rack',
        path_cluster_settings    = '/usr/local/sge-root/default/common/settings.sh'

             ): 
        pass



#LC = Launcher_E3()
LC = Launcher_Frankfurt()

LC.python                    = 'python2.6'
LC.notification_mail_address = 'lorenz@ke.tu-darmstadt.de'
LC.sync_executable_32        = '../rechner/programme/unison-2.27'
LC.verbose_sync              = 0


#LC.remote_base_dir     = None
#LC.remote_peewit_path = 'peewit/pw_base/python/'
LC.remote_sync_executable_32  = '../rechner/programme/unison-2.27'
LC.remote_sync_executable_64  = '../rechner/programme/unison-2.27_64bit'
LC.remote_python_paths_os  = []    
LC.exclude_from_sync       = [
                              'modules_extended_32/libsvm-2.86/python',
                              'modules_extended_64/libsvm-2.86/python',
                              ]




#LC.remote_head         = 'kerosin'
#LC.remote_head         = 'kerze'

## spec-element: (host-name, number of cores, architecture) 
LC.remote_nodes_spec   =  [#('kehrblech', 2, '32bit'),
                           ('kerosin',   4),# '64bit'),
                           #('kerze',     2),# '64bit'),
                           #('ketzer',    2, '32bit'),
                           #('key',       3),# '32bit'),                           
                          ]





### digest all this now

PK.digest_paths(AR,LC)

#import sys
#print "PP0"
#for p in sys.path: print p

### no matplotlib
#PathKeeper.use_pylab = 1


